import textwrap

import pytest

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient


def test_basic():
    client = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile

        class Pkg(ConanFile):

            def package_info(self):
                self.conf_info.define_path("tools.android:ndk_path", "MY-NDK!!!")
        """)
    client.save({"conanfile.py": conanfile})
    client.run("create . --name=android_ndk --version=1.0")

    consumer = textwrap.dedent("""
        from conan import ConanFile

        class Pkg(ConanFile):
            settings = "os", "compiler", "build_type", "arch"
            generators = "CMakeToolchain"
            build_requires = "android_ndk/1.0"

            def generate(self):
                self.output.info("NDK: %s" % self.conf.get("tools.android:ndk_path"))
        """)
    # CMakeToolchain needs compiler definition
    linux_profile = textwrap.dedent("""
        [settings]
        os = Linux
        arch = x86_64
        compiler = gcc
        compiler.version = 4.9
        compiler.libcxx = libstdc++
        """)
    android_profile = textwrap.dedent("""
        include(linux)
        [conf]
        tools.android:ndk_path=MY-SYSTEM-NDK!!!
        """)
    client.save({"conanfile.py": consumer,
                 "linux": linux_profile,
                 "android": android_profile}, clean_first=True)

    client.run("install . -pr=linux")
    assert "conanfile.py: NDK: MY-NDK!!!" in client.out

    client.run("install . -pr:b=default -pr:h=android")
    assert "conanfile.py: NDK: MY-SYSTEM-NDK!!!" in client.out


def test_basic_conf_through_cli():
    client = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile

        class Pkg(ConanFile):

            def package_info(self):
                self.output.info("NDK build: %s" % self.conf.get("tools.android:ndk_path"))
        """)
    client.save({"conanfile.py": conanfile})
    client.run("create . --name=android_ndk --version=1.0")

    consumer = textwrap.dedent("""
        from conan import ConanFile

        class Pkg(ConanFile):
            settings = "os", "compiler", "build_type", "arch"
            generators = "CMakeToolchain"
            build_requires = "android_ndk/1.0"

            def generate(self):
                self.output.info("NDK host: %s" % self.conf.get("tools.android:ndk_path"))
        """)
    # CMakeToolchain needs compiler definition
    linux_profile = textwrap.dedent("""
        [settings]
        os = Linux
        arch = x86_64
        compiler = gcc
        compiler.version = 4.9
        compiler.libcxx = libstdc++
        """)
    android_profile = textwrap.dedent("""
        include(linux)
        [conf]
        tools.android:ndk_path=MY-SYSTEM-NDK!!!
        """)
    client.save({"conanfile.py": consumer,
                 "linux": linux_profile,
                 "android": android_profile}, clean_first=True)
    client.run('install . -c:b=tools.android:ndk_path="MY-NDK!!!" '
               '-c:h=tools.android:ndk_path="MY-SYSTEM-NDK!!!" -pr:b=default -pr:h=android')
    assert "android_ndk/1.0: NDK build: MY-NDK!!!" in client.out
    assert "conanfile.py: NDK host: MY-SYSTEM-NDK!!!" in client.out


def test_declared_generators_get_conf():
    # https://github.com/conan-io/conan/issues/9571
    client = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        class Pkg(ConanFile):
            def package_info(self):
                self.conf_info.append("tools.cmake.cmaketoolchain:user_toolchain",
                                      "mytoolchain.cmake")
        """)
    client.save({"conanfile.py": conanfile})
    client.run("create . --name=mytool --version=1.0")

    consumer = textwrap.dedent("""
        from conan import ConanFile

        class Pkg(ConanFile):
            settings = "os", "compiler", "build_type", "arch"
            generators = "CMakeToolchain"
            build_requires = "mytool/1.0"
        """)
    client.save({"conanfile.py": consumer}, clean_first=True)
    client.run("install . -pr:b=default")
    toolchain = client.load("conan_toolchain.cmake")
    assert 'include("mytoolchain.cmake")' in toolchain

    consumer = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.cmake import CMakeToolchain

        class Pkg(ConanFile):
            settings = "os", "compiler", "build_type", "arch"
            build_requires = "mytool/1.0"

            def generate(self):
                CMakeToolchain(self).generate()
        """)
    client.save({"conanfile.py": consumer}, clean_first=True)
    client.run("install . -pr:b=default")
    toolchain = client.load("conan_toolchain.cmake")
    assert 'include("mytoolchain.cmake")' in toolchain


def test_propagate_conf_info():
    """ test we can use the conf_info to propagate information from the dependencies
    to the consumers. The propagation is explicit.
    TO DISCUSS: Should conf be aggregated always from all requires?
    TODO: Backport to Conan 1.X so UserInfo is not longer necessary in 1.X
    """
    # https://github.com/conan-io/conan/issues/9571
    client = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        class Pkg(ConanFile):
            def package_info(self):
                self.conf_info.define("user:myinfo1", "val1")
        """)
    client.save({"conanfile.py": conanfile})
    client.run("create . --name=dep1 --version=1.0")
    client.run("create . --name=dep2 --version=1.0")

    consumer = textwrap.dedent("""
        from conan import ConanFile

        class Pkg(ConanFile):
            requires = "dep1/1.0", "dep2/1.0"
            def generate(self):
                c1 = self.dependencies["dep1"].conf_info.get("user:myinfo1")
                c2 = self.dependencies["dep2"].conf_info.get("user:myinfo1")
                self.output.info("CONF1: {}".format(c1))
                self.output.info("CONF2: {}".format(c2))
        """)
    client.save({"conanfile.py": consumer}, clean_first=True)
    client.run("install . ")
    assert "conanfile.py: CONF1: val1" in client.out
    assert "conanfile.py: CONF2: val1" in client.out


def test_conf_transitive_tool():
    """
    # https://github.com/conan-io/conan/issues/14421
    app --(tool_requires)--> tool/0.1 -> lib/0.1 -(tool_require)-> libbuilder/0.1 -> zlib/0.1

    profile-host = Debug
    profile-build = Release
    """

    client = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        class Pkg(ConanFile):
            version = "0.1"
            settings = "build_type"
            {}
            def package_info(self):
                self.output.info(f"host: {{self.settings.build_type}}")
                self.output.info(f"build: {{self.settings_build.build_type}}")
                if self.settings_target is not None:
                    self.output.info(f"target: {{self.settings_target.build_type}}")
        """)
    client.save({"zlib/conanfile.py": conanfile.format(""),
                 "libbuilder/conanfile.py": conanfile.format("requires='zlib/0.1'"),
                 "lib/conanfile.py": conanfile.format("tool_requires='libbuilder/0.1'"),
                 "tool/conanfile.py": conanfile.format("requires='lib/0.1'"),
                 "app/conanfile.py": conanfile.format("tool_requires='tool/0.1'")})
    client.run("export zlib --name=zlib")
    client.run("export libbuilder --name=libbuilder")
    client.run("export lib --name=lib")
    client.run("export tool --name=tool")
    client.run("create app --name=app -s:b build_type=Release -s:h build_type=Debug --build=missing")
    for lib in "zlib", "libbuilder":
        assert f"{lib}/0.1: host: Release" in client.out
        assert f"{lib}/0.1: build: Release" in client.out
        assert f"{lib}/0.1: target: Release" in client.out  # used to create lib/0.1 that is Release!
    for lib in "lib", "tool":
        assert f"{lib}/0.1: host: Release" in client.out
        assert f"{lib}/0.1: build: Release" in client.out
        assert f"{lib}/0.1: target: Debug" in client.out  # used to create app/0.1 that is Debug!
    assert "app/0.1: host: Debug" in client.out
    assert "app/0.1: build: Release" in client.out


def test_conf_both_build_and_host():
    """
    # https://github.com/conan-io/conan/issues/14421
    app --(requires)-----> protobuf(lib-host-release) -(tool_require)-> tool/0.1
      \\- (tool_require)-> protobuf(protoc-build-release) -(tool_require) -> tool/0.1

    profile-host = Debug
    profile-build = Release
    """

    client = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        class Pkg(ConanFile):
            name = "tool"
            version = "0.1"
            def package_info(self):
                myvalue = str(self.settings_target.build_type)
                self.conf_info.define("user.team:myconf", myvalue)
        """)
    client.save({"conanfile.py": conanfile})
    client.run("create . --build-require")

    myprotobuf = textwrap.dedent("""
        from conan import ConanFile

        class Pkg(ConanFile):
            name = "myprotobuf"
            version = "0.1"
            settings = "build_type"
            tool_requires = "tool/0.1"

            def generate(self):
                self.output.info(f"MYCONF {self.context}: {self.conf.get('user.team:myconf')}")
        """)

    client.save({"myprotobuf/conanfile.py": myprotobuf,
                 "app/conanfile.py": GenConanfile().with_requires("myprotobuf/0.1")
                                                   .with_tool_requires("myprotobuf/0.1")},
                clean_first=True)
    client.run('export myprotobuf')
    client.run("install app --build=missing -s:h build_type=Debug -s:b build_type=Release")
    assert "myprotobuf/0.1: MYCONF build: Release" in client.out
    assert "myprotobuf/0.1: MYCONF host: Debug" in client.out


@pytest.mark.parametrize("conf", ['define("user.myorg:myconf", self.options.myopt)',
                                  'define("user.myorg:myconf", self.settings.os)',
                                  'append("user.myorg:myconf", self.settings.os)',
                                  'prepend("user.myorg:myconf", self.settings.os)'])
def test_error_bad_types(conf):
    c = TestClient(light=True)
    conanfile = textwrap.dedent(f"""
        from conan import ConanFile
        class Pkg(ConanFile):
            name = "tool"
            version = "0.1"
            settings = "os"
            options = {{"myopt": [True, False]}}
            default_options = {{"myopt": True}}
            def package_info(self):
                self.conf_info.{conf}
        """)
    c.save({"conanfile.py": conanfile})
    c.run("create .", assert_error=True)
    assert "ERROR: tool/0.1: Error in package_info() method, line 10"
    assert f'self.conf_info.{conf}' in c.out
    assert "Invalid 'conf' type, please use Python types (int, str, ...)" in c.out


@pytest.mark.parametrize("conf", ["tools.myorg.myconf", "user.myorg.myconf"])
def test_error_missing_colon_define(conf):
    c = TestClient(light=True)
    conanfile = textwrap.dedent(f"""
        from conan import ConanFile
        class Pkg(ConanFile):
            name = "tool"
            version = "0.1"
            def package_info(self):
                self.conf_info.define("{conf}", 42)
        """)
    c.save({"conanfile.py": conanfile})
    c.run("create .", assert_error=True)
    assert "ERROR: tool/0.1: Error in package_info() method:" in c.out
    if conf == "tools.myorg.myconf":
        assert "[conf] 'tools.myorg.myconf' does not exist in configuration" in c.out
    else:
        assert f"User conf '{conf}' invalid format, not 'user.org.group:conf'" in c.out


def test_error_missing_colon_consume():
    c = TestClient(light=True)
    conanfile = textwrap.dedent(f"""
        from conan import ConanFile
        class Pkg(ConanFile):
            def generate(self):
                self.conf.get("user.myorg.myconf")
        """)
    c.save({"conanfile.py": conanfile})
    c.run("install .", assert_error=True)
    assert "User conf 'user.myorg.myconf' invalid format, not 'user.org.group:conf'" in c.out



import os
import platform
import pytest

from conan.tools.microsoft.subsystems import subsystem_path
from conan.internal.subsystems import get_cased_path
from conan.test.utils.test_files import temp_folder
from conan.internal.util.files import mkdir


class TestGetCasedPath:
    @pytest.mark.skipif(platform.system() != "Windows", reason="Requires Windows")
    def test_case_existing(self):
        folder = get_cased_path(temp_folder())
        p1 = os.path.join(folder, "MyFolder", "Subfolder")
        mkdir(p1)

        assert p1 == get_cased_path(p1)  # Idempotent
        assert p1 == get_cased_path(os.path.join(folder, "myfolder", "subfolder"))

    def test_case_not_existing(self):
        current_dir = get_cased_path(os.getcwd())
        non_existing_path = os.path.join(current_dir, "this", "Path", "does", "NOT", "Exists")
        p = get_cased_path(non_existing_path)  # If not exists from the root, returns as is
        assert p == non_existing_path

    @pytest.mark.skipif(platform.system() != "Windows", reason="Requires Windows")
    def test_case_partial_exists(self):
        folder = get_cased_path(temp_folder())
        p1 = os.path.join(folder, "MyFolder", "Subfolder")
        mkdir(p1)

        non_existing_path = os.path.join(folder, "myfolder", "subfolder", "not-existing")
        # The first part of the path will be properly cased.
        assert os.path.join(p1, "not-existing") == get_cased_path(non_existing_path)


class TestUnixPath:

    def test_none(self):
        assert None == subsystem_path(None, path=None)

    def test_msys_path(self):
        assert '/c/windows/system32' == subsystem_path("msys2", 'C:\\Windows\\System32')

    def test_cygwin_path(self):
        assert '/cygdrive/c/windows/system32' == subsystem_path("cygwin",
                                                               'C:\\Windows\\System32')

        # another drive
        assert '/cygdrive/d/work' == subsystem_path("cygwin", "D:\\work")

        # path inside the cygwin
        assert '/home/.conan' == subsystem_path("cygwin", '/home/.conan')
        assert '/dev/null' == subsystem_path("cygwin", '/dev/null')

        # relative paths
        assert './configure' == subsystem_path("cygwin", './configure')
        assert '../configure' == subsystem_path("cygwin", '../configure')
        assert 'source_subfolder/configure' == subsystem_path("cygwin", 'source_subfolder/configure')

        assert './configure' == subsystem_path("cygwin", '.\\configure')
        assert '../configure' == subsystem_path("cygwin", '..\\configure')
        assert 'source_subfolder/configure' == subsystem_path("cygwin", 'source_subfolder\\configure')

        # already with cygdrive
        assert '/cygdrive/c/conan' == subsystem_path("cygwin", '/cygdrive/c/conan')

        # UNC (file share)
        assert '//server/share' == subsystem_path("cygwin", "\\\\SERVER\\Share")

        # long path
        assert '/cygdrive/c/windows/system32' == subsystem_path("cygwin", '\\\\?\\C:\\Windows\\System32')

    def test_wsl_path(self):
        assert '/mnt/c/Windows/System32' == subsystem_path("wsl", 'C:\\Windows\\System32')



import pytest

from conan.api.model import ListPattern
from conan.errors import ConanException


@pytest.mark.parametrize("pattern, result",
                         [("*",                  ("*", "latest", None, "latest")),
                          ("zlib/1.2.11",        ("zlib/1.2.11", "latest", None, "latest")),
                          ("zlib/1.2.11#rev1",   ("zlib/1.2.11", "rev1", None, "latest")),
                          ("zlib/1.2.11:pid1",   ("zlib/1.2.11", "latest", "pid1", "latest"))])
def test_cli_pattern_matching(pattern, result):
    pattern = ListPattern(pattern)
    assert result == (pattern.ref, pattern.rrev, pattern.package_id, pattern.prev)


def test_list_pattern():
    with pytest.raises(ConanException) as e:
        ListPattern("*:*", only_recipe=True)
    assert "Do not specify 'package_id' with 'only-recipe'" in str(e.value)



import json
import os
import textwrap

import pytest

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.test_files import temp_folder
from conan.test.utils.tools import TestClient
from conan.internal.util.files import mkdir, save, save_files


@pytest.fixture(scope="module")
def c3i_folder():
    folder = temp_folder()
    recipes_folder = os.path.join(folder, "recipes")
    zlib_config = textwrap.dedent("""
        versions:
          "1.2.8":
            folder: all
          "1.2.11":
            folder: all
        """)
    zlib = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.files import load
        class Zlib(ConanFile):
            name = "zlib"
            exports_sources = "*"
            def build(self):
                self.output.info(f"CONANDATA: {self.conan_data}")
                self.output.info(f"BUILDING: {load(self, 'file.h')}")
            """)
    save_files(recipes_folder,
               {"zlib/config.yml": zlib_config,
                "zlib/all/conanfile.py": zlib,
                "zlib/all/conandata.yml": "",
                "zlib/all/file.h": "//myheader"})
    mkdir(os.path.join(recipes_folder, "openssl", "1.X"))
    mkdir(os.path.join(recipes_folder, "openssl", "2.X"))
    save(os.path.join(recipes_folder, "openssl", "config.yml"), textwrap.dedent("""
        versions:
          "1.0":
            folder: "1.X"
          "1.1":
            folder: "1.X"
          "2.0":
            folder: "2.X"
        """))
    save(os.path.join(recipes_folder, "openssl", "1.X", "conanfile.py"),
         str(GenConanfile().with_require("zlib/1.2.8")))
    save(os.path.join(recipes_folder, "openssl", "2.X", "conanfile.py"),
         str(GenConanfile().with_require("zlib/1.2.11")))
    mkdir(os.path.join(recipes_folder, "libcurl", "all"))
    save(os.path.join(recipes_folder, "libcurl", "config.yml"), textwrap.dedent("""
            versions:
              "1.0":
                folder: "all"
            """))
    save(os.path.join(recipes_folder, "libcurl", "all", "conanfile.py"),
         str(GenConanfile().with_require("openssl/2.0")))

    save(os.path.join(recipes_folder, ".DS_Store", "foo"), "")

    return folder


class TestSearchList:
    def test_basic_search(self, c3i_folder):
        client = TestClient(light=True)
        client.run(f"remote add local '{c3i_folder}' --type=local-recipes-index")  # Keep --type test
        assert "WARN" not in client.out  # Make sure it does not complain about url
        client.run("search *")
        assert textwrap.dedent("""\
            local
              libcurl
                libcurl/1.0
              openssl
                openssl/1.0
                openssl/1.1
                openssl/2.0
              zlib
                zlib/1.2.8
                zlib/1.2.11
            """) in client.out

    def test_list_refs(self, c3i_folder):
        client = TestClient(light=True)
        client.run(f"remote add local '{c3i_folder}'")
        client.run("list *#* -r=local --format=json")
        listjson = json.loads(client.stdout)
        revs = listjson["local"]["libcurl/1.0"]["revisions"]
        assert len(revs) == 1 and "e468388f0e4e098d5b62ad68979aebd5" in revs
        revs = listjson["local"]["openssl/1.0"]["revisions"]
        assert len(revs) == 1 and "b35ffb31b6d5a9d8af39f5de3cf4fd63" in revs
        revs = listjson["local"]["openssl/1.1"]["revisions"]
        assert len(revs) == 1 and "b35ffb31b6d5a9d8af39f5de3cf4fd63" in revs
        revs = listjson["local"]["openssl/2.0"]["revisions"]
        assert len(revs) == 1 and "e50e871efca149f160fa6354c8534449" in revs
        revs = listjson["local"]["zlib/1.2.8"]["revisions"]
        assert len(revs) == 1 and "6f5c31bb1219e9393743d1fbf2ee1b52" in revs
        revs = listjson["local"]["zlib/1.2.11"]["revisions"]
        assert len(revs) == 1 and "6f5c31bb1219e9393743d1fbf2ee1b52" in revs

    def test_list_revisions_notfound(self, c3i_folder):
        client = TestClient(light=True)
        client.run(f"remote add local '{c3i_folder}'")
        client.run("list potato/1.0#* -r=local")
        # More like remotes than cache
        assert "ERROR: Recipe not found: 'potato/1.0'"

    def test_list_rrevs(self, c3i_folder):
        client = TestClient(light=True)
        client.run(f"remote add local '{c3i_folder}'")
        client.run("list libcurl/1.0#* -r=local --format=json")
        listjson = json.loads(client.stdout)
        revs = listjson["local"]["libcurl/1.0"]["revisions"]
        assert len(revs) == 1 and "e468388f0e4e098d5b62ad68979aebd5" in revs

    def test_list_binaries(self, c3i_folder):
        client = TestClient(light=True)
        client.run(f"remote add local '{c3i_folder}'")
        client.run("list libcurl/1.0:* -r=local --format=json")
        listjson = json.loads(client.stdout)
        rev = listjson["local"]["libcurl/1.0"]["revisions"]["e468388f0e4e098d5b62ad68979aebd5"]
        assert rev["packages"] == {}


class TestInstall:
    def test_install(self, c3i_folder):
        c = TestClient(light=True)
        c.run(f"remote add local '{c3i_folder}'")
        c.run("install --requires=libcurl/1.0 --build missing")
        assert "zlib/1.2.11: CONANDATA: {}" in c.out
        assert "zlib/1.2.11: BUILDING: //myheader" in c.out
        bins = {"libcurl/1.0": ("aa69c1e1e39a18fe70001688213dbb7ada95f890", "Build"),
                "openssl/2.0": ("594ed0eb2e9dfcc60607438924c35871514e6c2a", "Build"),
                "zlib/1.2.11": ("da39a3ee5e6b4b0d3255bfef95601890afd80709", "Build")}
        c.assert_listed_binary(bins)

        # Already installed in the cache
        c.run("install --requires=libcurl/1.0")
        assert "zlib/1.2.11: Already installed!" in c.out
        assert "openssl/2.0: Already installed!" in c.out
        assert "libcurl/1.0: Already installed!" in c.out

        # Update doesn't fail, but doesn't update revision time
        c.run("install --requires libcurl/1.0 --update")
        bins = {"libcurl/1.0": "Cache (Updated date) (local)",
                "openssl/2.0": "Cache (Updated date) (local)",
                "zlib/1.2.11": "Cache (Updated date) (local)"}

        c.assert_listed_require(bins)
        assert "zlib/1.2.11: Already installed!" in c.out
        assert "openssl/2.0: Already installed!" in c.out
        assert "libcurl/1.0: Already installed!" in c.out

        # Doing local changes creates a new revision
        # New recipe revision for the zlib library
        save(os.path.join(c3i_folder, "recipes", "zlib", "all", "conanfile.py"),
             str(GenConanfile()) + "\n")
        c.run("install --requires=libcurl/1.0 --build missing --update")
        # it is updated
        assert "zlib/1.2.11#dd82451a95902c89bb66a2b980c72de5 - Updated (local)" in c.out

    def test_install_with_exported_files(self):
        folder = temp_folder()
        recipes_folder = os.path.join(folder, "recipes")
        boost_config = textwrap.dedent("""
            versions:
              "1.0":
                folder: all
            """)
        boost = textwrap.dedent("""
            import os
            from conan.tools.files import load
            from conan import ConanFile
            class Boost(ConanFile):
                name = "boost"
                version = "1.0"
                exports = "*"
                def source(self):
                    myfile = os.path.join(self.recipe_folder, "dependencies", "myfile.json")
                    self.output.info(load(self, myfile))
                """)
        deps_json = '{"potato": 42}'
        save_files(recipes_folder,
                   {"boost/config.yml": boost_config,
                    "boost/all/conanfile.py": boost,
                    "boost/all/dependencies/myfile.json": deps_json})
        c = TestClient(light=True)
        c.run(f"remote add local '{folder}'")
        c.run("install --requires=boost/[*] --build missing")
        assert 'boost/1.0: {"potato": 42}' in c.out

    def test_trim_conandata_yaml(self):
        folder = temp_folder()
        recipes_folder = os.path.join(folder, "recipes")
        config = textwrap.dedent("""
            versions:
              "1.0":
                folder: all
            """)
        conandata = textwrap.dedent("""\
            sources:
              "1.0":
                url:
                sha256: "ff0ba4c292013dbc27530b3a81e1f9a813cd39de01ca5e0f8bf355702efa593e"
            patches:
              "1.0":
                - patch_file: "patches/1.3/0001-fix-cmake.patch"
            """)
        save_files(recipes_folder,
                   {"pkg/config.yml": config,
                    "pkg/all/conanfile.py": str(GenConanfile("pkg")),
                    "pkg/all/conandata.yml": conandata})
        c = TestClient(light=True)
        c.run(f"remote add local '{folder}'")
        c.run("install --requires=pkg/1.0 --build missing -vvv")
        assert "pkg/1.0#86b609916bbdfe63c579f034ad0edfe7" in c.out

        # User modifies conandata.yml to add new version
        new_conandata = textwrap.dedent("""\
            sources:
              "1.0":
                url:
                sha256: "ff0ba4c292013dbc27530b3a81e1f9a813cd39de01ca5e0f8bf355702efa593e"
              "1.1":
                url:
            patches:
              "1.0":
                - patch_file: "patches/1.3/0001-fix-cmake.patch"
            """)
        save_files(recipes_folder, {"pkg/all/conandata.yml": new_conandata})
        c.run("install --requires=pkg/1.0 --build missing --update -vvv")
        assert "pkg/1.0#86b609916bbdfe63c579f034ad0edfe7" in c.out

    def test_export_patches(self):
        folder = temp_folder()
        recipes_folder = os.path.join(folder, "recipes")
        zlib_config = textwrap.dedent("""
            versions:
              "0.1":
                folder: all
            """)
        zlib = textwrap.dedent("""
            from conan import ConanFile
            from conan.tools.files import export_conandata_patches, apply_conandata_patches
            class Zlib(ConanFile):
                name = "zlib"
                exports_sources = "*.cpp"
                def export_sources(self):
                    export_conandata_patches(self)

                def source(self):
                    apply_conandata_patches(self)
                    """)
        conandata_yml = textwrap.dedent("""\
            versions:
              "0.1":
            patches:
              "0.1":
                - patch_file: "patches/patch1"
                    """)
        patch = textwrap.dedent("""\
            --- a/main.cpp
            +++ b/main.cpp
            @@ -0,0 +1 @@
            +hello
            """)
        save_files(recipes_folder,
                   {"zlib/config.yml": zlib_config,
                    "zlib/all/conanfile.py": zlib,
                    "zlib/all/conandata.yml": conandata_yml,
                    "zlib/all/patches/patch1": patch,
                    "zlib/all/main.cpp": "\n"})
        client = TestClient(light=True)
        client.run(f"remote add local '{folder}'")
        client.run("install --requires=zlib/0.1 --build=missing -vv")
        assert "zlib/0.1: Copied 1 file: patch1" in client.out
        assert "zlib/0.1: Apply patch (file): patches/patch1" in client.out

    def test_export_user_channel(self):
        folder = temp_folder()
        recipes_folder = os.path.join(folder, "recipes")
        zlib_config = textwrap.dedent("""
            versions:
              "0.1":
                folder: all
            """)
        zlib = GenConanfile("zlib").with_class_attribute("user='myuser'")\
                                   .with_class_attribute("channel='mychannel'")
        conandata_yml = textwrap.dedent("""\
            versions:
              "0.1":
            """)
        save_files(recipes_folder, {"zlib/config.yml": zlib_config,
                                    "zlib/all/conanfile.py": str(zlib),
                                    "zlib/all/conandata.yml": conandata_yml})
        client = TestClient()
        client.run(f"remote add local '{folder}'")
        client.run("install --requires=zlib/0.1@myuser/mychannel --build=missing")
        assert "zlib/0.1@myuser/mychannel:" in client.out
        client.run("list * -r=local")
        assert "zlib/0.1@myuser/mychannel" in client.out


class TestRestrictedOperations:
    def test_upload(self):
        folder = temp_folder()
        c3i_folder = os.path.join(folder, "recipes")
        mkdir(c3i_folder)
        c = TestClient(light=True)
        c.run(f"remote add local '{c3i_folder}'")
        c.save({"conanfile.py": GenConanfile("pkg", "0.1")})
        c.run("create .")
        c.run("upload pkg/0.1 -r=local", assert_error=True)
        assert "ERROR: Remote local-recipes-index 'local' doesn't support upload" in c.out


class TestErrorsUx:
    def test_errors(self):
        folder = temp_folder()
        recipes_folder = os.path.join(folder, "recipes")
        zlib_config = textwrap.dedent("""
            versions:
              "1.2.11":
                folder: all
            """)
        zlib = textwrap.dedent("""
            class Zlib(ConanFile):
                name = "zlib"
                """)
        save_files(recipes_folder,
                   {"zlib/config.yml": zlib_config,
                    "zlib/all/conanfile.py": zlib})
        c = TestClient(light=True)
        c.run(f"remote add local '{folder}'")
        c.run("install --requires=zlib/[*] --build missing", assert_error=True)
        assert "NameError: name 'ConanFile' is not defined" in c.out

    def test_require_revision(self):
        # https://github.com/conan-io/conan/issues/17814
        folder = temp_folder()
        recipes_folder = os.path.join(folder, "recipes")
        zlib_config = textwrap.dedent("""
           versions:
             "1.2.11":
               folder: all
           """)
        save_files(recipes_folder,
                   {"zlib/config.yml": zlib_config,
                    "zlib/all/conanfile.py": str(GenConanfile("zlib"))})
        c = TestClient(light=True)
        c.run(f"remote add local '{folder}'")
        c.run("install --requires=zlib/1.2.11#rev1", assert_error=True)
        assert "A specific revision 'zlib/1.2.11#rev1' was requested" in c.out

    def test_no_user_channel(self):
        # https://github.com/conan-io/conan/issues/18142
        folder = temp_folder()
        recipes_folder = os.path.join(folder, "recipes")
        zlib_config = textwrap.dedent("""
          versions:
            "0.1":
              folder: all
          """)
        zlib = GenConanfile("zlib")
        conandata_yml = textwrap.dedent("""\
          versions:
            "0.1":
          """)
        save_files(recipes_folder, {"zlib/config.yml": zlib_config,
                                    "zlib/all/conanfile.py": str(zlib),
                                    "zlib/all/conandata.yml": conandata_yml})
        c = TestClient()
        c.run(f"remote add local '{folder}'")
        c.run("install --requires=zlib/0.1@myuser/mychannel", assert_error=True)
        assert "ERROR: Package 'zlib/0.1@myuser/mychannel' not resolved" in c.out

        c = TestClient(default_server_user=True)
        c.save({"conanfile.py": GenConanfile("zlib", "0.1")})
        c.run("create . --user=myuser --channel=mychannel")
        c.run("upload * -r=default -c")
        c.run(f"remote add local '{folder}' --index=0")
        c.run("install --requires=zlib/0.1@myuser/mychannel")
        c.assert_listed_require({"zlib/0.1@myuser/mychannel": "Cache"})

        # Force resolving in remotes
        c.run("remove * -c")
        c.run("install --requires=zlib/0.1@myuser/mychannel")
        c.assert_listed_require({"zlib/0.1@myuser/mychannel": "Downloaded (default)"})

    def test_errors_missing_folder(self):
        folder = temp_folder()
        repo = os.path.join(folder, "repo")
        mkdir(repo)
        c = TestClient(light=True)
        c.run(f"remote add local '{repo}'")
        # shutil.rmtree(repo)
        c.run("install --requires=zlib/[*] --build missing", assert_error=True)
        assert "Cannot connect to 'local-recipes-index' repository, missing 'recipes'" in c.out


class TestPythonRequires:
    @pytest.fixture(scope="class")
    def c3i_pyrequires_folder(self):
        folder = temp_folder()
        recipes_folder = os.path.join(folder, "recipes")
        config = textwrap.dedent("""
            versions:
              "1.0":
                folder: all
            """)
        pkg = str(GenConanfile("pkg").with_python_requires("pyreq/1.0"))
        save_files(recipes_folder,
                   {"pkg/config.yml": config,
                    "pkg/all/conanfile.py": pkg,
                    "pyreq/config.yml": config,
                    "pyreq/all/conanfile.py": str(GenConanfile("pyreq", "1.0"))})
        return folder

    def test_install(self, c3i_pyrequires_folder):
        c = TestClient(light=True)
        c.run(f"remote add local '{c3i_pyrequires_folder}'")
        c.run("list * -r=local")
        assert "pyreq/1.0" in c.out
        assert "pkg/1.0" in c.out
        c.run("install --requires=pkg/1.0 --build missing -vvv")
        assert "pyreq/1.0#a0d63ca853edefa33582a24a1bb3c75f - Downloaded (local)" in c.out
        assert "pkg/1.0: Created package" in c.out


class TestUserChannel:
    @pytest.fixture(scope="class")
    def c3i_user_channel_folder(self):
        folder = temp_folder()
        recipes_folder = os.path.join(folder, "recipes")
        config = textwrap.dedent("""
                versions:
                  "1.0":
                    folder: all
                  "2.0":
                    folder: other
                """)
        pkg = str(GenConanfile("pkg").with_class_attribute("user='myuser'")
                                     .with_class_attribute("channel='mychannel'"))
        save_files(recipes_folder,
                   {"pkg/config.yml": config,
                    "pkg/all/conanfile.py": pkg,
                    "pkg/other/conanfile.py": str(GenConanfile("pkg"))})
        return folder

    def test_user_channel_requirement(self, c3i_user_channel_folder):
        tc = TestClient(light=True)
        tc.run(f"remote add local '{c3i_user_channel_folder}'")
        tc.run("graph info --requires=pkg/[*]@myuser/mychannel")
        assert "Version range '*' from requirement 'pkg/[*]@myuser/mychannel'" not in tc.out
        assert "pkg/[*]@myuser/mychannel: pkg/1.0@myuser/mychannel" in tc.out
        assert "pkg/1.0@myuser/mychannel#0b23a5938afb0457079e41aac8991595 - Downloaded (local)" in tc.out

    @pytest.mark.parametrize("user_channel", ["@foo/bar", "@foo"])
    def test_user_channel_requirement_no_match(self, c3i_user_channel_folder, user_channel):
        tc = TestClient(light=True)
        tc.run(f"remote add local '{c3i_user_channel_folder}'")
        tc.run(f"graph info --requires=pkg/[*]{user_channel}", assert_error=True)
        assert f"Version range '*' from requirement 'pkg/[*]{user_channel}' required by 'None' could not be resolved." in tc.out

    def test_user_channel_requirement_only_at(self, c3i_user_channel_folder):
        tc = TestClient(light=True)
        tc.run(f"remote add local '{c3i_user_channel_folder}'")
        tc.run(f"graph info --requires=pkg/[*]@")
        assert f"pkg/[*]: pkg/2.0" in tc.out

        tc.run(f"graph info --requires=pkg/[<2]@", assert_error=True)
        assert f" Package 'pkg/[<2]' not resolved" in tc.out

        tc.run(f"graph info --requires=pkg/[<2]", assert_error=True)
        assert f" Package 'pkg/[<2]' not resolved" in tc.out


class TestResetRemote:
    def test_resetting_remote_error(self):
        # https://github.com/conan-io/conan/issues/18371
        folder = temp_folder()
        recipes_folder = os.path.join(folder, "recipes")
        zlib_config = textwrap.dedent("""
            versions:
              "1.2.11":
                folder: all
            """)
        zlib = textwrap.dedent("""
            from conan import ConanFile
            from conan.tools.files import load
            class Zlib(ConanFile):
                name = "zlib"
                exports_sources = "*"
                """)
        save_files(recipes_folder,
                   {"zlib/config.yml": zlib_config,
                    "zlib/all/conanfile.py": zlib,
                    "zlib/all/conandata.yml": "",
                    "zlib/all/file.h": "//myheader"})

        client = TestClient(light=True)
        client.run(f"remote add local '{folder}'")
        client.run("graph info --requires=zlib/[*]")

        # This second --force destroys the previous remote database
        client.run(f"remote add local '{folder}' --force")
        client.run("install --requires=zlib/[*] --build=missing")
        # It doesn't fail or crash anymore

    def test_changing_revisions(self):
        # https://github.com/conan-io/conan/issues/18371
        folder = temp_folder()
        recipes_folder = os.path.join(folder, "recipes")
        zlib_config = textwrap.dedent("""
            versions:
              "1.2.11":
                folder: all
            """)
        zlib = textwrap.dedent("""
            from conan import ConanFile
            from conan.tools.files import load
            class Zlib(ConanFile):
                name = "zlib"
                exports_sources = "*"
                def build(self):
                    self.output.info(f"BUILDING: {load(self, 'file.h')}")
            """)
        save_files(recipes_folder,
                   {"zlib/config.yml": zlib_config,
                    "zlib/all/conanfile.py": zlib,
                    "zlib/all/conandata.yml": "",
                    "zlib/all/file.h": "//myheader"})

        c = TestClient(light=True)
        c.run(f"remote add local '{folder}'")
        c.run("graph info --requires=zlib/[*]")
        rev1 = "bd69839cb4c933336fceb32302aaf91f"

        # Modify zlib code
        save_files(recipes_folder, {"zlib/all/file.h": "//myheader 222"})
        c.run("graph info --requires=zlib/[*] --update")
        rev2 = "c912566276abca17d2fb5fb6fc957852"

        c.run(f"remote add local '{folder}' --force")
        c.run(f"install --requires=zlib/1.2.11#{rev2} --build=missing")  # works
        c.run(f"install --requires=zlib/1.2.11#{rev1} --build=missing", assert_error=True)
        assert ("WARN: A specific revision 'zlib/1.2.11#bd69839cb4c933336fceb32302aaf91f' was "
                "requested, but it doesn't match the current available revision in source") in c.out
        assert ("ERROR: The 'zlib/1.2.11' package has 'exports_sources' but sources "
                "not found in local cache") in c.out



import fnmatch
import os
import sys
import time
from threading import Lock

import colorama
from colorama import Fore, Style

from conan.errors import ConanException

LEVEL_QUIET = 80  # -q
LEVEL_ERROR = 70  # Errors
LEVEL_WARNING = 60  # Warnings
LEVEL_NOTICE = 50  # Important messages to attract user attention.
LEVEL_STATUS = 40  # Default - The main interesting messages that users might be interested in.
LEVEL_VERBOSE = 30  # -v  Detailed informational messages.
LEVEL_DEBUG = 20  # -vv Closely related to internal implementation details
LEVEL_TRACE = 10  # -vvv Fine-grained messages with very low-level implementation details


class Color:
    """ Wrapper around colorama colors that are undefined in importing
    """
    RED = Fore.RED  # @UndefinedVariable
    WHITE = Fore.WHITE  # @UndefinedVariable
    CYAN = Fore.CYAN  # @UndefinedVariable
    GREEN = Fore.GREEN  # @UndefinedVariable
    MAGENTA = Fore.MAGENTA  # @UndefinedVariable
    BLUE = Fore.BLUE  # @UndefinedVariable
    YELLOW = Fore.YELLOW  # @UndefinedVariable
    BLACK = Fore.BLACK  # @UndefinedVariable

    BRIGHT_RED = Style.BRIGHT + Fore.RED  # @UndefinedVariable
    BRIGHT_BLUE = Style.BRIGHT + Fore.BLUE  # @UndefinedVariable
    BRIGHT_YELLOW = Style.BRIGHT + Fore.YELLOW  # @UndefinedVariable
    BRIGHT_GREEN = Style.BRIGHT + Fore.GREEN  # @UndefinedVariable
    BRIGHT_CYAN = Style.BRIGHT + Fore.CYAN  # @UndefinedVariable
    BRIGHT_WHITE = Style.BRIGHT + Fore.WHITE  # @UndefinedVariable
    BRIGHT_MAGENTA = Style.BRIGHT + Fore.MAGENTA  # @UndefinedVariable


if os.environ.get("CONAN_COLOR_DARK"):
    Color.WHITE = Fore.BLACK
    Color.CYAN = Fore.BLUE
    Color.YELLOW = Fore.MAGENTA
    Color.BRIGHT_WHITE = Fore.BLACK
    Color.BRIGHT_CYAN = Fore.BLUE
    Color.BRIGHT_YELLOW = Fore.MAGENTA
    Color.BRIGHT_GREEN = Fore.GREEN


def init_colorama(stream):
    import colorama
    if _color_enabled(stream):
        if os.getenv("CLICOLOR_FORCE", "0") != "0":
            # Otherwise it is not really forced if colorama doesn't feel it
            colorama.init(strip=False, convert=False)
        else:
            colorama.init()


def _color_enabled(stream):
    """
    NO_COLOR: No colors

    https://no-color.org/

    Command-line software which adds ANSI color to its output by default should check for the
    presence of a NO_COLOR environment variable that, when present (**regardless of its value**),
    prevents the addition of ANSI color.

    CLICOLOR_FORCE: Force color

    https://bixense.com/clicolors/
    """

    if os.getenv("CLICOLOR_FORCE", "0") != "0":
        # CLICOLOR_FORCE != 0, ANSI colors should be enabled no matter what.
        return True

    if os.getenv("NO_COLOR") is not None:
        return False
    return hasattr(stream, "isatty") and stream.isatty()


class ConanOutput:
    """ A singleton class to handle output messages in Conan.

    Recipes should only access this class through the ``self.output`` attribute of the recipe,
    but custom commands or tools can instantiate it directly, where doing so for each message is
    a valid practice.

    It provides methods to write messages at different levels of verbosity, such as debug, info,
    warning, and error. The output level can be controlled by the user through command-line options
    or environment variables.

    The output methods return the instance itself, so different methods can be chained together.
    """
    # Singleton
    _conan_output_level = LEVEL_STATUS
    _silent_warn_tags = []
    _warnings_as_errors = []
    lock = Lock()

    def __init__(self, scope: str = ""):
        """ Initialize the ConanOutput instance.

        :parameter scope: A string that represents the scope of the output. This is usually the
            reference of the recipe being executed, like ``pkg/1.0@user/channel`` and is prefixed
            to the output messages. If not provided, it defaults to an empty string.
        """
        self.stream = sys.stderr
        self._scope = scope
        # FIXME:  This is needed because in testing we are redirecting the sys.stderr to a buffer
        #         stream to capture it, so colorama is not there to strip the color bytes
        self._color = _color_enabled(self.stream)

    @classmethod
    def define_silence_warnings(cls, warnings):
        cls._silent_warn_tags = warnings

    @classmethod
    def set_warnings_as_errors(cls, value):
        cls._warnings_as_errors = value

    @classmethod
    def get_output_level(cls):
        return cls._conan_output_level

    @classmethod
    def set_output_level(cls, level):
        cls._conan_output_level = level

    @classmethod
    def valid_log_levels(cls):
        return {"quiet": LEVEL_QUIET,  # -vquiet 80
            "error": LEVEL_ERROR,  # -verror 70
            "warning": LEVEL_WARNING,  # -vwaring 60
            "notice": LEVEL_NOTICE,  # -vnotice 50
            "status": LEVEL_STATUS,  # -vstatus 40
            None: LEVEL_VERBOSE,  # -v 30
            "verbose": LEVEL_VERBOSE,  # -vverbose 30
            "debug": LEVEL_DEBUG,  # -vdebug 20
            "v": LEVEL_DEBUG,  # -vv 20
            "trace": LEVEL_TRACE,  # -vtrace 10
            "vv": LEVEL_TRACE  # -vvv 10
        }

    @classmethod
    def define_log_level(cls, v):
        env_level = os.getenv("CONAN_LOG_LEVEL")
        v = env_level or v
        levels = cls.valid_log_levels()
        try:
            level = levels[v]
        except KeyError:
            msg = " defined in CONAN_LOG_LEVEL environment variable" if env_level else ""
            vals = "quiet, error, warning, notice, status, verbose, debug(v), trace(vv)"
            raise ConanException(f"Invalid argument '-v{v}'{msg}.\nAllowed values: {vals}")
        else:
            cls.set_output_level(level)

    @classmethod
    def level_allowed(cls, level):
        return cls._conan_output_level <= level

    @property
    def color(self):
        return self._color

    @property
    def scope(self):
        return self._scope

    @scope.setter
    def scope(self, out_scope):
        self._scope = out_scope

    @property
    def is_terminal(self):
        return hasattr(self.stream, "isatty") and self.stream.isatty()

    def writeln(self, data, fg=None, bg=None):
        return self.write(data, fg, bg, newline=True)

    def write(self, data, fg=None, bg=None, newline=False):
        if self._conan_output_level > LEVEL_NOTICE:
            return self
        if self._color and (fg or bg):
            data = "%s%s%s%s" % (fg or '', bg or '', data, Style.RESET_ALL)

        if newline:
            data = "%s\n" % data

        with self.lock:
            self.stream.write(data)
            self.stream.flush()

        return self

    def box(self, msg: str):
        """ Draw a box around the message, useful for important messages"""
        color = Color.BRIGHT_GREEN
        self.writeln("\n**************************************************", fg=color)
        self.writeln(f'*{msg: ^48}*', fg=color)
        self.writeln(f"**************************************************\n", fg=color)
        return self

    def login_msg(self, msg, newline=False):
        # unconditional to the error level, this has to show always
        self._write_message(msg, newline=newline)
        return self

    def _write_message(self, msg, fg=None, bg=None, newline=True):
        if isinstance(msg, dict):
            # For traces we can receive a dict already, we try to transform then into more natural
            # text
            msg = ", ".join([f"{k}: {v}" for k, v in msg.items()])
            msg = "=> {}".format(msg)
            # msg = json.dumps(msg, sort_keys=True, default=json_encoder)

        if self._scope:
            if self._color:
                ret = f"{fg or ''}{bg or ''}{self._scope}: {msg}{Style.RESET_ALL}"
            else:
                ret = f"{self._scope}: {msg}"
        else:
            if self._color:
                ret = f"{fg or ''}{bg or ''}{msg}{Style.RESET_ALL}"
            else:
                ret = msg

        if newline:
            ret = f"{ret}\n"

        with self.lock:
            self.stream.write(ret)
            self.stream.flush()

    def trace(self, msg: str):
        """ This is the most extreme level of detail.

        Trace messages log every little step the system takes, including function entries and exits,
        variable changes, and other very specific events.

        This message won't be printed unless the user has set the log level to trace
        (e.g., using the ``-vvv`` option in the command line).

        It’s used when full visibility of everything happening in the system is required,
        but should be used carefully due to the large amount of information it can generate."""
        if self._conan_output_level <= LEVEL_TRACE:
            self._write_message(msg, fg=Color.BLUE)
        return self

    def debug(self, msg: str, fg: str = Color.MAGENTA, bg: str = None):
        """ With a high level of detail, it is mainly used for debugging code.

        This message won't be printed unless the user has set the log level to debug
        (e.g., using the ``-vv`` option in the command line).

        These messages provide useful information for developers, such as variable values
        or execution flow details, to trace errors or analyze the program's behavior."""
        if self._conan_output_level <= LEVEL_DEBUG:
            self._write_message(msg, fg=fg, bg=bg)
        return self

    def verbose(self, msg: str, fg: str = None, bg: str = None):
        """ Displays additional and detailed information that, while not critical,
        can be useful for better understanding how the system is working.

        This message won't be printed unless the user has set the log level to verbose
        (e.g., using the ``-v`` option in the command line).

        It’s appropriate for gaining more context without overloading the logs with
        excessive detail. Useful when more clarity is needed than a simple info."""
        if self._conan_output_level <= LEVEL_VERBOSE:
            self._write_message(msg, fg=fg, bg=bg)
        return self

    def status(self, msg: str, fg: str = None, bg: str = None, newline: bool = True):
        """ Provides general information about the system or ongoing operations.

        Info messages are basic and used to inform about common events,
        like the start or completion of processes, without implying specific problems or achievements."""
        if self._conan_output_level <= LEVEL_STATUS:
            self._write_message(msg, fg=fg, bg=bg, newline=newline)
        return self

    info = status

    def title(self, msg: str):
        """ Draws a title around the message, useful for important messages"""
        if self._conan_output_level <= LEVEL_NOTICE:
            self._write_message("\n======== {} ========".format(msg),
                                fg=Color.BRIGHT_MAGENTA)
        return self

    def subtitle(self, msg: str):
        """ Draws a subtitle around the message, useful for important messages"""
        if self._conan_output_level <= LEVEL_NOTICE:
            self._write_message("\n-------- {} --------".format(msg),
                                fg=Color.BRIGHT_MAGENTA)
        return self

    def highlight(self, msg: str):
        """ Marks or emphasizes important events or processes that need to stand out but don’t necessarily
        indicate success or error.

        These messages draw attention to key points that may be relevant for the user or administrator."""
        if self._conan_output_level <= LEVEL_NOTICE:
            self._write_message(msg, fg=Color.BRIGHT_MAGENTA)
        return self

    def success(self, msg: str):
        """ Shows that an operation has been completed successfully.

        This type of message is useful to confirm that key processes or tasks have finished correctly,
        which is essential for good application monitoring."""
        if self._conan_output_level <= LEVEL_NOTICE:
            self._write_message(msg, fg=Color.BRIGHT_GREEN)
        return self

    @staticmethod
    def _warn_tag_matches(warn_tag, patterns):
        lookup_tag = warn_tag or "unknown"
        return any(fnmatch.fnmatch(lookup_tag, pattern) for pattern in patterns)

    def warning(self, msg: str, warn_tag: str = None):
        """ Highlights a potential issue that, while not stopping the system,
        could cause problems in the future or under certain conditions.

        Warnings signal abnormal situations that should be
        reviewed but don’t necessarily cause an immediate halt in operations.
        Notice that if the tag matches the pattern in the ``core:warnings_as_errors`` configuration,
        and is not skipped, this will be upgraded to an error, and raise an exception
        when the output is printed, so that the error does not pass unnoticed."""
        _treat_as_error = self._warn_tag_matches(warn_tag, self._warnings_as_errors)
        if (self._conan_output_level <= LEVEL_WARNING or
                (_treat_as_error and self._conan_output_level <= LEVEL_ERROR)):
            if self._warn_tag_matches(warn_tag, self._silent_warn_tags):
                return self
            warn_tag_msg = "" if warn_tag is None else f"{warn_tag}: "
            output = f"{warn_tag_msg}{msg}"

            if _treat_as_error:
                self.error(output)
            else:
                self._write_message(f"WARN: {output}", Color.YELLOW)
        return self

    def error(self, msg: str, error_type: str = None):
        """ Indicates that a serious issue has occurred that prevents the system
        or application from continuing to function correctly.

        Typically, this represents a failure in the normal flow of execution,
        such as a service crash or a critical exception.
        Notice that if the user has set the ``core:warnings_as_errors`` configuration,
        this will raise an exception when the output is printed,
        so that the error does not pass unnoticed."""
        if self._warnings_as_errors and error_type != "exception":
            raise ConanException(msg)
        if self._conan_output_level <= LEVEL_ERROR:
            self._write_message("ERROR: {}".format(msg), Color.RED)
        return self

    def flush(self):
        self.stream.flush()


def cli_out_write(data, fg=None, bg=None, endline="\n", indentation=0):
    """
    Output to be used by formatters to dump information to stdout
    """
    if (fg or bg) and _color_enabled(sys.stdout):  # need color
        data = f"{' ' * indentation}{fg or ''}{bg or ''}{data}{Style.RESET_ALL}{endline}"
        sys.stdout.write(data)
    else:
        data = f"{' ' * indentation}{data}{endline}"
        if sys.stdout.isatty():
            # https://github.com/conan-io/conan/issues/17245 avoid colorama crash and overhead
            # skip deinit/reinit if stdout is not a TTY to preserve redirected output to file
            colorama.deinit()
            sys.stdout.write(data)
            colorama.reinit()
        else:
            sys.stdout.write(data)


class TimedOutput:
    def __init__(self, interval, out=None, msg_format=None):
        self._interval = interval
        self._msg_format = msg_format
        self._t = time.time()
        self._out = out or ConanOutput()

    def info(self, msg, *args, **kwargs):
        t = time.time()
        if t - self._t > self._interval:
            self._t = t
            if self._msg_format:
                msg = self._msg_format(msg, *args, **kwargs)
            self._out.info(msg)



import textwrap

import pytest

from conan.test.utils.tools import TestClient, GenConanfile


PKG_ID_1 = "47b42eaf657374a3d040394f03961b66c53bda5e"
PKG_ID_2 = "8b7006bf91e5b52cc1ac24a7a4d9c326ee954bb2"


class TestPythonRequiresPackageID:

    @pytest.fixture(autouse=True)
    def set_up(self):
        client = TestClient()
        client.save({"conanfile.py": GenConanfile()})
        client.run("export . --name=tool --version=1.1.1")
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            class Pkg(ConanFile):
                python_requires ="tool/[*]"
            """)
        client2 = TestClient(cache_folder=client.cache_folder)
        client2.save({"conanfile.py": conanfile})
        self.client = client
        self.client2 = client2

    def test_default(self):
        self.client2.run("create . --name=pkg --version=0.1")
        assert "tool/1.1.1" in self.client2.out
        pkg_id = "170e82ef3a6bf0bbcda5033467ab9d7805b11d0b"
        self.client2.assert_listed_binary({"pkg/0.1": (pkg_id, "Build")})

        self.client.run("export . --name=tool --version=1.1.2")
        self.client2.run("create . --name=pkg --version=0.1")
        assert "tool/1.1.2" in self.client2.out
        self.client2.assert_listed_binary({"pkg/0.1": (pkg_id,
                                                       "Build")})

        # With a minor change, it fires a rebuild
        self.client.run("export . --name=tool --version=1.2.0")
        self.client2.run("create . --name=pkg --version=0.1")
        assert "tool/1.2.0" in self.client2.out
        self.client2.assert_listed_binary({"pkg/0.1": ("5eb1e7ea93fdd67fe3c3b166d240844648ba2b7a",
                                                       "Build")})

    def test_change_mode_conf(self):
        # change the policy in conan.conf
        self.client2.save_home({"global.conf": "core.package_id:default_python_mode=patch_mode"})
        self.client2.run("create . --name=pkg --version=0.1")
        assert "tool/1.1.1" in self.client2.out
        self.client2.assert_listed_binary({"pkg/0.1": (PKG_ID_1, "Build")})

        # with a patch change, new ID
        self.client.run("export . --name=tool --version=1.1.2")
        self.client2.run("create . --name=pkg --version=0.1")
        assert "tool/1.1.2" in self.client2.out
        self.client2.assert_listed_binary({"pkg/0.1": (PKG_ID_2, "Build")})

    def test_unrelated_conf(self):
        # change the policy in conan.conf
        self.client2.save_home({"global.conf": "core.package_id:default_python_mode=unrelated_mode"})
        self.client2.run("create . --name=pkg --version=0.1")
        assert "tool/1.1.1" in self.client2.out
        pkg_id = "da39a3ee5e6b4b0d3255bfef95601890afd80709"
        self.client2.assert_listed_binary({"pkg/0.1": (pkg_id, "Build")})

        # with any change the package id doesn't change
        self.client.run("export . --name=tool --version=1.1.2")
        self.client2.run("create . --name=pkg --version=0.1 --build missing")
        assert "tool/1.1.2" in self.client2.out
        self.client2.assert_listed_binary({"pkg/0.1": (pkg_id, "Cache")})

    def test_change_mode_package_id(self):
        # change the policy in package_id
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            class Pkg(ConanFile):
                python_requires ="tool/[*]"
                def package_id(self):
                    self.info.python_requires.patch_mode()
            """)
        self.client2.save({"conanfile.py": conanfile})
        self.client2.run("create . --name=pkg --version=0.1")
        assert "tool/1.1.1" in self.client2.out
        self.client2.assert_listed_binary({"pkg/0.1": (PKG_ID_1, "Build")})

        # with a patch change, new ID
        self.client.run("export . --name=tool --version=1.1.2")
        self.client2.run("create . --name=pkg --version=0.1")
        assert "tool/1.1.2" in self.client2.out
        self.client2.assert_listed_binary({"pkg/0.1": (PKG_ID_2, "Build")})


def test_python_requires_for_build_requires():
    client = TestClient()
    client.save_home({"global.conf": "core.package_id:default_python_mode=full_version_mode"})
    client.save({"conanfile.py": GenConanfile()})
    client.run("create . --name=tool --version=1.1.1")

    client2 = TestClient(cache_folder=client.cache_folder)
    client2.save({"conanfile.py": GenConanfile().with_python_requires("tool/[>=0.0]"),
                 "myprofile": "[tool_requires]\ntool/[>=0.0]\n"})

    client2.run("create . --name=pkg --version=0.1 -pr=myprofile")
    assert "tool/1.1.1" in client2.out
    assert f"pkg/0.1: Package '{PKG_ID_1}' created" in client2.out

    client.run("create . --name=tool --version=1.1.2")
    client2.run("install --requires=pkg/0.1@ -pr=myprofile", assert_error=True)
    assert f"ERROR: Missing binary: pkg/0.1:{PKG_ID_2}" in client2.out
    assert "tool/1.1.2" in client2.out
    assert "tool/1.1.1" not in client2.out

    client2.run("create . --name=pkg --version=0.1 -pr=myprofile")
    # assert "pkg/0.1: Applying build-requirement: tool/1.1.2", client2.out)
    assert f"pkg/0.1: Package '{PKG_ID_2}' created" in client2.out


class TestPythonRequiresHeaderOnly:
    def test_header_only(self):
        c = TestClient(light=True)
        pkg = textwrap.dedent("""\
            from conan import ConanFile
            class Pkg(ConanFile):
                name = "pkg"
                version = "0.1"
                python_requires = "tool/[*]"
                def package_id(self):
                    self.info.clear()
                """)
        c.save({"tool/conanfile.py": GenConanfile("tool"),
                "pkg/conanfile.py": pkg})
        c.run("create tool --version=1.0")
        c.run("create pkg")
        pkgid = c.created_package_id("pkg/0.1")
        c.run("create tool --version=1.2")
        c.run("install --requires=pkg/0.1")
        c.assert_listed_binary({"pkg/0.1": (pkgid, "Cache")})

    def test_header_only_implements(self):
        c = TestClient(light=True)
        pkg = textwrap.dedent("""\
            from conan import ConanFile
            class Pkg(ConanFile):
                name = "pkg"
                version = "0.1"
                python_requires = "tool/[*]"
                package_type = "header-library"
                implements = ["auto_header_only"]
                """)
        c.save({"tool/conanfile.py": GenConanfile("tool"),
                "pkg/conanfile.py": pkg})
        c.run("create tool --version=1.0")
        c.run("create pkg")
        pkgid = c.created_package_id("pkg/0.1")
        c.run("create tool --version=1.2")
        c.run("install --requires=pkg/0.1")
        c.assert_listed_binary({"pkg/0.1": (pkgid, "Cache")})


@pytest.mark.parametrize("mode, pkg_id",
                         [("unrelated_mode", "da39a3ee5e6b4b0d3255bfef95601890afd80709"),
                          ("semver_mode", "1f0070b00ccebfec93dc90854a163c7af229f587"),
                          ("patch_mode", "b9ca872dfd5b48f5f1f69d66f0950fc35469d0cd"),
                          ("minor_mode", "c53bd9e48dd09ceeaa1bb425830490d8b243e39c"),
                          ("major_mode", "331c17383dcdf37f79bc2b86fa55ac56afdc6fec"),
                          ("full_version_mode", "1f0070b00ccebfec93dc90854a163c7af229f587"),
                          ("revision_mode", "0071dd0296afa0db533e21f924273485c87f0d32"),
                          ("full_mode", "0071dd0296afa0db533e21f924273485c87f0d32")])
def test_modes(mode, pkg_id):
    c = TestClient(light=True)
    c.save_home({"global.conf": f"core.package_id:default_python_mode={mode}"})
    c.save({"dep/conanfile.py": GenConanfile("dep", "0.1.1.1"),
            "pkg/conanfile.py": GenConanfile("pkg", "0.1").with_python_requires("dep/[*]")})
    c.run("create dep")
    c.run("create pkg")
    pkgid = c.created_package_id("pkg/0.1")
    assert pkgid == pkg_id



import textwrap

from conan.test.utils.tools import TestClient, GenConanfile


class TestLoopDetection:

    def test_transitive_loop(self):
        client = TestClient(light=True)
        client.save({
            'pkg1.py': GenConanfile().with_require('pkg2/0.1@lasote/stable'),
            'pkg2.py': GenConanfile().with_require('pkg3/0.1@lasote/stable'),
            'pkg3.py': GenConanfile().with_require('pkg1/0.1@lasote/stable'),
        })
        client.run('export pkg1.py --name=pkg1 --version=0.1 --user=lasote --channel=stable')
        client.run('export pkg2.py --name=pkg2 --version=0.1 --user=lasote --channel=stable')
        client.run('export pkg3.py --name=pkg3 --version=0.1 --user=lasote --channel=stable')

        client.run("install --requires=pkg3/0.1@lasote/stable --build='*'", assert_error=True)
        # TODO: Complete with better diagnostics
        assert "ERROR: There is a cycle/loop in the graph" in client.out

    def test_self_loop(self):
        client = TestClient(light=True)
        client.save({'pkg1.py': GenConanfile().with_require('pkg1/0.1@lasote/stable'), })
        client.run('export pkg1.py --name=pkg1 --version=0.1 --user=lasote --channel=stable')
        client.run("install --requires=pkg1/0.1@lasote/stable --build='*'", assert_error=True)
        assert "ERROR: There is a cycle/loop in the graph" in client.out


def test_install_order_infinite_loop():
    c = TestClient(light=True)
    c.save({"fmt/conanfile.py": GenConanfile("fmt", "1.0"),
            "tool/conanfile.py": GenConanfile("tool", "1.0").with_requires("fmt/1.0"),
            "tool_profile": "[tool_requires]\n!tool/*: tool/1.0"})
    c.run("export fmt")
    c.run("export tool")
    c.run("install tool -pr:h=tool_profile -b=missing",
          assert_error=True)
    assert "ERROR: There is a loop in the graph" in c.out
    assert "fmt/1.0 (Build) -> ['tool/1.0']" in c.out
    assert "tool/1.0 (Build) -> ['fmt/1.0']" in c.out

    # Graph build-order fails in the same way
    c.run("graph build-order tool -pr:h=tool_profile -b=missing",
          assert_error=True)
    assert "ERROR: There is a loop in the graph" in c.out
    assert "fmt/1.0 (Build) -> ['tool/1.0']" in c.out
    assert "tool/1.0 (Build) -> ['fmt/1.0']" in c.out

    c.run("graph build-order tool -pr:h=tool_profile --order-by=configuration -b=missing",
          assert_error=True)
    assert "ERROR: There is a loop in the graph" in c.out
    assert "fmt/1.0:da39a3ee5e6b4b0d3255bfef95601890afd80709 (Build) -> " \
           "['tool/1.0:044d18636d2b7da86d3aa46a2aabf1400db525b1']" in c.out
    assert "tool/1.0:044d18636d2b7da86d3aa46a2aabf1400db525b1 (Build) -> " \
           "['fmt/1.0:da39a3ee5e6b4b0d3255bfef95601890afd80709']" in c.out


def test_build_require_undetected_loop():
    c = TestClient(light=True)
    conanfile = textwrap.dedent("""
        from conan import ConanFile

        class DemoRecipe(ConanFile):
            name = "test"
            version = "1.0"

            def requirements(self):
                self.requires("cmake/3.31.6", build=True)

            def build_requirements(self):
                self.tool_requires("cmake/3.31.6", options={
                    "bad": True # this doesn't need to be a valid option
                })

            def generate(self):
                self.output.info(f"NUM DEPS: {len(self.dependencies.items())}")
        """)
    c.save({"cmake/conanfile.py": GenConanfile("cmake", "3.31.6").with_shared_option(True),
            "app/conanfile.py": conanfile})

    c.run("create cmake")
    c.run("install app")
    # It doesn't hang, and it sees correctly just 1 dependency
    assert "conanfile.py (test/1.0): NUM DEPS: 1" in c.out



import json
import os
import base64

from jinja2 import Template

from conan.api.output import cli_out_write
from conan.cli.formatters.report.diff_html import diff_html


def _generate_json(result):
    diff_text = result["diff"]
    src_prefix = result["src_prefix"]
    dst_prefix = result["dst_prefix"]
    ret = {}
    current_filename = None
    for line in diff_text.splitlines():
        if line.startswith("diff --git "):
            src_filename, dst_filename = _get_filenames(line, src_prefix, dst_prefix)
            current_filename = src_filename
            ret[current_filename] = [line]
        else:
            ret[current_filename].append(line)
    return ret


def _get_filenames(line, src_prefix, dst_prefix):
    """
    Extracts the source and destination filenames from a diff line.
    """
    src_index = line.find(src_prefix)
    dst_index = line.find(dst_prefix)

    if src_index == -1 or dst_index == -1:
        return None, None

    src_filename = line[src_index + len(src_prefix) - 1:dst_index - 1].strip()
    dst_filename = line[dst_index + len(dst_prefix) - 1:].strip()

    return src_filename, dst_filename


def _render_diff(content, template, template_folder, **kwargs):
    from conan import __version__
    template = Template(template, autoescape=True)

    def _safe_filename(filename):
        # Calculate base64 of the filename
        return base64.b64encode(filename.encode(), altchars=b'-_').decode()

    def _remove_prefixes(line):
        return line.replace(kwargs["src_prefix"][:-1], "").replace(kwargs["dst_prefix"][:-1], "")

    def _replace_cache_paths(line):
        return line.replace(kwargs["old_cache_path"], "(old)").replace(kwargs["new_cache_path"], "(new)")

    def _replace_paths(line):
        return _remove_prefixes(_replace_cache_paths(line))

    def _extract_header(diff_lines):
        # Header ends at the first occurrence of +++ line,
        # and it can be at most 10 lines long
        for i, line in enumerate(diff_lines[:10]):
            if line.startswith("+++ "):
                return diff_lines[:i + 1]
        return diff_lines[:10]

    def _parse_header_is_deleted(header_contents):
        return ("+++ /dev/null" in header_contents
                or any("deleted file mode" in line for line in header_contents))

    def _parse_header_rename_to(header_contents):
        if not any("similarity index" in line for line in header_contents):
            return None
        for line in header_contents:
            if line.startswith("rename to "):
                return line[len("rename to "):]
        return None

    per_folder = {"folders": {}, "files": {}}
    for file in content:
        header = _extract_header(content[file])
        renamed_to = _parse_header_rename_to(header)
        replaced_path = _replace_paths(renamed_to or file)
        replaced_file = replaced_path.replace("(old)", "").replace("(new)", "").replace("\\", "/")
        bits = replaced_file.split("/")[1:]
        cur = per_folder
        for folder in bits[:-1]:
            cur = cur["folders"].setdefault(folder, {"folders": {}, "files": {}})
        filename = bits[-1]
        cur["files"][filename] = {"filename": file,  # This is file so renamed use old name
                                  "is_new": "(new)" in replaced_path,
                                  "is_deleted": _parse_header_is_deleted(header),
                                  "renamed_to": renamed_to,
                                  "relative_path": replaced_path}

    def flatten_empty_folders(current_node):
        for folder_data in current_node["folders"].values():
            flatten_empty_folders(folder_data)

        promoted_folders = {}

        # The list here is important to avoid modifying the dict while iterating
        for folder_name, folder_data in list(current_node["folders"].items()):
            if not folder_data["files"]:
                for sub_folder_name, sub_folder_data in folder_data['folders'].items():
                    new_key = os.path.join(folder_name, sub_folder_name)
                    promoted_folders[new_key] = sub_folder_data

                del current_node["folders"][folder_name]

        current_node["folders"].update(promoted_folders)

    flatten_empty_folders(per_folder)

    # Now sort each folder and file recursively
    def sort_folders_and_files(node):
        node["folders"] = dict(sorted(node["folders"].items()))
        node["files"] = dict(sorted(node["files"].items(), key=lambda x: x[0].lower()))
        for folder_data in node["folders"].values():
            sort_folders_and_files(folder_data)
    sort_folders_and_files(per_folder)

    return template.render(content=content,
                           per_folder=per_folder,
                           base_template_path=template_folder, version=__version__,
                           safe_filename=_safe_filename,
                           replace_paths=_replace_paths,
                           replace_cache_paths=_replace_cache_paths,
                           remove_prefixes=_remove_prefixes,
                           **kwargs)


def format_diff_html(result):
    conan_api = result["conan_api"]

    template_folder = os.path.join(conan_api.cache_folder, "templates")
    user_template = os.path.join(template_folder, "diff.html")
    template = diff_html
    if os.path.isfile(user_template):
        with open(user_template, 'r', encoding="utf-8", newline="") as handle:
            template = handle.read()

    content = _generate_json(result)

    cli_out_write(_render_diff(content, template, template_folder,
                               old_reference=result["old_export_ref"],
                               new_reference=result["new_export_ref"],
                               old_cache_path=result["old_cache_path"],
                               new_cache_path=result["new_cache_path"],
                               src_prefix=result["src_prefix"],
                               dst_prefix=result["dst_prefix"]))


def format_diff_txt(result):
    diff_text = result["diff"]
    cli_out_write(diff_text)


def format_diff_json(result):
    cli_out_write(json.dumps(_generate_json(result), indent=2))



import json
import os
import textwrap

import pytest

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient


def test_profile_path():
    c = TestClient()
    c.run("profile path default")
    assert "default" in c.out


def test_profile_path_missing():
    c = TestClient()
    c.run("profile path notexisting", assert_error=True)
    assert "ERROR: Profile not found: notexisting" in c.out


def test_ignore_paths_when_listing_profiles():
    c = TestClient()
    ignore_path = '.DS_Store'

    # just in case
    os.makedirs(c.paths.profiles_path, exist_ok=True)
    # This a "touch" equivalent
    open(os.path.join(c.paths.profiles_path, '.DS_Store'), 'w').close()
    os.utime(os.path.join(c.paths.profiles_path, ".DS_Store"))

    c.run("profile list")

    assert ignore_path not in c.out


def test_shorthand_syntax():
    tc = TestClient()
    tc.save({"profile": "[conf]\nuser:profile=True"})
    tc.run("profile show -h")
    assert "[-pr:b" in tc.out
    assert "[-pr:h" in tc.out
    assert "[-pr:a" in tc.out

    tc.run(
        "profile show -o:a=both_options=True -pr:a=profile -s:a=os=WindowsCE -s:a=os.platform=conan -c:a=user.conf:cli=True -f=json")

    out = json.loads(tc.stdout)
    assert out == {'build': {'build_env': '',
                             'conf': {'user.conf:cli': True, 'user:profile': True},
                             'options': {'both_options': 'True'},
                             'package_settings': {},
                             'settings': {'os': 'WindowsCE', 'os.platform': 'conan'},
                             'tool_requires': {}},
                   'host': {'build_env': '',
                            'conf': {'user.conf:cli': True, 'user:profile': True},
                            'options': {'both_options': 'True'},
                            'package_settings': {},
                            'settings': {'os': 'WindowsCE', 'os.platform': 'conan'},
                            'tool_requires': {}}}

    tc.save({"pre": textwrap.dedent("""
            [settings]
            os=Linux
            compiler=gcc
            compiler.version=11
            """),
             "mid": textwrap.dedent("""
            [settings]
            compiler=clang
            compiler.version=14
            """),
             "post": textwrap.dedent("""
            [settings]
            compiler.version=13
            """)})

    tc.run("profile show -pr:a=pre -pr:a=mid -pr:a=post -f=json")
    out = json.loads(tc.stdout)
    assert out == {'build': {'build_env': '',
                             'conf': {},
                             'options': {},
                             'package_settings': {},
                             'settings': {'compiler': 'clang',
                                          'compiler.version': '13',
                                          'os': 'Linux'},
                             'tool_requires': {}},
                   'host': {'build_env': '',
                            'conf': {},
                            'options': {},
                            'package_settings': {},
                            'settings': {'compiler': 'clang',
                                         'compiler.version': '13',
                                         'os': 'Linux'},
                            'tool_requires': {}}}

    tc.run("profile show -pr:a=pre -pr:h=post -f=json")
    out = json.loads(tc.stdout)
    assert out == {'build': {'build_env': '',
                             'conf': {},
                             'options': {},
                             'package_settings': {},
                             'settings': {'compiler': 'gcc',
                                          'compiler.version': '11',
                                          'os': 'Linux'},
                             'tool_requires': {}},
                   'host': {'build_env': '',
                            'conf': {},
                            'options': {},
                            'package_settings': {},
                            'settings': {'compiler': 'gcc',
                                         'compiler.version': '13',
                                         'os': 'Linux'},
                            'tool_requires': {}}}

    tc.run("profile show -pr:a=pre -o:b foo=False -o:a foo=True -o:h foo=False -f=json")
    out = json.loads(tc.stdout)
    assert out == {'build': {'build_env': '',
                             'conf': {},
                             'options': {'foo': 'True'},
                             'package_settings': {},
                             'settings': {'compiler': 'gcc',
                                          'compiler.version': '11',
                                          'os': 'Linux'},
                             'tool_requires': {}},
                   'host': {'build_env': '',
                            'conf': {},
                            'options': {'foo': 'False'},
                            'package_settings': {},
                            'settings': {'compiler': 'gcc',
                                         'compiler.version': '11',
                                         'os': 'Linux'},
                            'tool_requires': {}}}
    tc.run("profile show -o:shared=True", assert_error=True)
    assert 'Invalid empty package name in options. Use a pattern like `mypkg/*:shared`' in tc.out


def test_profile_show_json():
    c = TestClient()
    c.save({"myprofilewin": "[settings]\nos=Windows\n"
                            "[tool_requires]\nmytool/*:mytool/1.0\n"
                            "[conf]\nuser.conf:value=42\nlibiconv/*:tools.env.virtualenv:powershell=False\n"
                            "[options]\n*:myoption=True\n"
                            "[replace_requires]\ncmake/*: cmake/3.29.0\n"
                            "[platform_requires]\ncmake/3.29.0\n",
            "myprofilelinux": "[settings]\nos=Linux"})
    c.run("profile show -pr:b=myprofilewin -pr:h=myprofilelinux --format=json")

    profile = json.loads(c.stdout)
    assert profile["host"]["settings"] == {"os": "Linux"}

    assert profile["build"]["settings"] == {"os": "Windows"}
    # Check that tool_requires are properly serialized in json format
    # https://github.com/conan-io/conan/issues/15183
    assert profile["build"]["tool_requires"] == {'mytool/*': ["mytool/1.0"]}


@pytest.mark.parametrize("new_global", [None, 0, 1, True, False, ""])
def test_settings_serialize(new_global):
    tc = TestClient(light=True)
    settings_user = "new_global: [null, 0, 1, True, False, '']"
    tc.save_home({"settings_user.yml": settings_user})
    tc.save({"conanfile.py": GenConanfile("foo", "1.0").with_settings("new_global")})

    s = f"-s new_global={new_global}" if new_global is not None else ""
    tc.run(f"create . {s}")
    if new_global is None:
        assert "settings: new_global=" not in tc.out
    else:
        assert f"settings: new_global={new_global}" in tc.out



import os
import platform
import textwrap

from conan.test.assets.premake import gen_premake5
from conan.test.utils.mocks import ConanFileMock
from conan.tools.files.files import replace_in_file, rmdir
import pytest

from conan.test.utils.tools import TestClient
from conan.test.assets.sources import gen_function_cpp


@pytest.mark.skipif(platform.machine() != "x86_64",
                    reason="Premake Legacy generator only supports x86_64 machines")
@pytest.mark.tool("premake")
def test_premake_legacy(matrix_client):
    c = matrix_client
    conanfile = textwrap.dedent("""
        import os
        from conan import ConanFile
        from conan.tools.premake import Premake
        from conan.tools.microsoft import MSBuild
        class Pkg(ConanFile):
            settings = "os", "compiler", "build_type", "arch"
            requires = "matrix/1.0"
            generators = "PremakeDeps", "VCVars"
            def build(self):
                p = Premake(self)
                p.configure()
                build_type = str(self.settings.build_type)
                if self.settings.os == "Windows":
                    msbuild = MSBuild(self)
                    msbuild.build("HelloWorld.sln")
                else:
                    self.run(f"make config={build_type.lower()}_x86_64")
                p = os.path.join(self.build_folder, "bin", build_type, "HelloWorld")
                self.run(f'"{p}"')
        """)
    premake = textwrap.dedent("""
        -- premake5.lua

        include('conandeps.premake5.lua')

        workspace "HelloWorld"
           conan_setup()
           configurations { "Debug", "Release" }
           platforms { "x86_64" }

        project "HelloWorld"
           kind "ConsoleApp"
           language "C++"
           targetdir "bin/%{cfg.buildcfg}"

           files { "**.h", "**.cpp" }

           filter "configurations:Debug"
              defines { "DEBUG" }
              symbols "On"

           filter "configurations:Release"
              defines { "NDEBUG" }
              optimize "On"

           filter "platforms:x86_64"
              architecture "x86_64"
          """)
    c.save({"conanfile.py": conanfile,
            "premake5.lua": premake,
            "main.cpp": gen_function_cpp(name="main", includes=["matrix"], calls=["matrix"])})
    c.run("build .")
    assert "main: Release!" in c.out
    assert "matrix/1.0: Hello World Release!" in c.out
    if platform.system() == "Windows":
        assert "main _M_X64 defined" in c.out
    else:
        assert "main __x86_64__ defined" in c.out
    c.run("build . -s build_type=Debug --build=missing")
    assert "main: Debug!" in c.out
    assert "matrix/1.0: Hello World Debug!" in c.out


@pytest.mark.tool("premake")
def test_premake_new_generator():
    c = TestClient()
    c.run("new premake_lib -d name=lib -d version=0.1 -o lib")
    c.run("create lib")
    c.run("new premake_exe -d name=example -d requires=lib/0.1 -d version=1.0 -o exe")
    c.run("create exe")

    assert "example/1.0 (test package): RUN: example" in c.out
    assert "lib/0.1: Hello World Release!" in c.out
    assert "example/1.0: Hello World Release!" in c.out


@pytest.mark.tool("premake")
def test_premake_shared_lib():
    c = TestClient()
    c.run("new premake_lib -d name=lib -d version=0.1 -o lib")
    c.run("create lib -o '&:shared=True'")
    assert "lib/0.1: package(): Packaged 1 '.so' file: liblib.so" in c.out
    assert "lib/0.1: package(): Packaged 1 '.a' file: liblib.a" not in c.out


@pytest.mark.tool("premake")
@pytest.mark.parametrize("transitive_libs", [True, False])
def test_premake_components(transitive_libs):
    c = TestClient()
    c.run("new premake_lib -d name=liba -d version=1.0 -o liba")

    libb_premake = gen_premake5(
        workspace="Libb",
        includedirs=[".", "libb1", "libb2"],
        projects=[
            {"name": "libb1", "files": ["libb1/*.h", "libb1/*.cpp"], "kind": "StaticLib"},
            {"name": "libb2", "files": ["libb2/*.h", "libb2/*.cpp"], "kind": "StaticLib",
             "links": ["libb1"]}
        ]
    )
    libb_conanfile = textwrap.dedent("""
        import os
        from conan import ConanFile
        from conan.tools.files import copy
        from conan.tools.layout import basic_layout
        from conan.tools.premake import Premake


        class libbRecipe(ConanFile):
            name = "libb"
            version = "1.0"
            package_type = "static-library"
            settings = "os", "compiler", "build_type", "arch"
            exports_sources = "premake5.lua", "libb1/*", "libb2/*"
            generators= "PremakeDeps", "PremakeToolchain"

            def layout(self):
                basic_layout(self)

            def requirements(self):
                self.requires("liba/1.0", transitive_libs=%s)

            def build(self):
                premake = Premake(self)
                premake.configure()
                premake.build(workspace="Libb")

            def package(self):
                for lib in ("libb1", "libb2"):
                    copy(self, "*.h", os.path.join(self.source_folder, lib),
                         os.path.join(self.package_folder, "include", lib))

                for pattern in ("*.lib", "*.a", "*.so*", "*.dylib"):
                    copy(self, pattern, os.path.join(self.build_folder, "bin"),
                         os.path.join(self.package_folder, "lib"))
                copy(self, "*.dll", os.path.join(self.build_folder, "bin"),
                     os.path.join(self.package_folder, "bin"))

            def package_info(self):
                self.cpp_info.components["libb1"].libs = ["libb1"]
                self.cpp_info.components["libb1"].requires = ["liba::liba"]

                self.cpp_info.components["libb2"].libs = ["libb2"]
                self.cpp_info.components["libb2"].requires = ["libb1"]
    """)

    libb_h = textwrap.dedent("""
        #pragma once
        #include <string>
        void libb%s();
    """)

    libb_cpp = textwrap.dedent("""
        #include <iostream>
        #include "libb%s.h"
        #include "liba.h"
        void libb%s(){
            liba();
            std::cout << "libb%s/1.0" << std::endl;
        }
    """)

    c.save(
        {
            "libb/premake5.lua": libb_premake,
            "libb/conanfile.py": libb_conanfile % transitive_libs,
            "libb/libb1/libb1.h": libb_h % "1",
            "libb/libb1/libb1.cpp": libb_cpp % ("1", "1", "1"),
            "libb/libb2/libb2.h": libb_h % "2",
            "libb/libb2/libb2.cpp": libb_cpp % ("2", "2", "2"),
        }
    )
    # Create a consumer application which depends on libb
    c.run("new premake_exe -d name=consumer -d version=1.0 -o consumer -d requires=libb/1.0")
    # Adapt includes and usage of libb in the consumer application
    replace_in_file(ConanFileMock(),
                    os.path.join(c.current_folder, "consumer", "src", "consumer.cpp"),
                    '#include "libb.h"', '#include "libb1/libb1.h"\n#include "libb2/libb2.h"')
    replace_in_file(ConanFileMock(),
                    os.path.join(c.current_folder, "consumer", "src", "consumer.cpp"),
                    'libb()', 'libb1();libb2()')

    c.run("create liba")
    c.run("create libb")
    # If transitive_libs is false, consumer will not compile because it will not find liba
    c.run("build consumer", assert_error=not transitive_libs)


@pytest.mark.tool("premake")
def test_transitive_headers_not_public(transitive_libraries):
    c = transitive_libraries

    main = gen_function_cpp(name="main", includes=["engine"], calls=["engine"])
    premake5 = gen_premake5(
        workspace="Consumer",
        projects=[
            {"name": "consumer", "files": ["src/main.cpp"], "kind": "ConsoleApp"}
        ],
    )

    conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.layout import basic_layout
        from conan.tools.premake import Premake

        class ConsumerRecipe(ConanFile):
            name = "consumer"
            version = "1.0"
            package_type = "application"
            settings = "os", "compiler", "build_type", "arch"
            exports_sources = "*"
            generators= "PremakeDeps", "PremakeToolchain"

            def layout(self):
                basic_layout(self)

            def requirements(self):
                self.requires("engine/1.0")

            def build(self):
                premake = Premake(self)
                premake.configure()
                premake.build(workspace="Consumer")
    """)

    c.save({"src/main.cpp": main,
            "premake5.lua": premake5,
            "conanfile.py": conanfile
            })
    # Test it builds successfully
    c.run("build .")

    # Test including a transitive header: it should fail as engine does not require matrix with
    # transitive_headers enabled
    main = gen_function_cpp(name="main", includes=["engine", "matrix"], calls=["engine"])
    c.save({"src/main.cpp": main})
    rmdir(ConanFileMock(), os.path.join(c.current_folder, "build-release"))
    c.run("build .", assert_error=True)
    # Error should be about not finding matrix

@pytest.mark.tool("premake")
def test_premake_custom_configuration(transitive_libraries):
    c = transitive_libraries
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.layout import basic_layout
        from conan.tools.premake import Premake, PremakeDeps, PremakeToolchain

        class ConsumerRecipe(ConanFile):
            name = "consumer"
            version = "1.0"
            package_type = "application"
            settings = "os", "compiler", "build_type", "arch"
            options = {"sanitizer": [True, False]}
            default_options = {"sanitizer": False}
            exports_sources = "*"

            def layout(self):
                basic_layout(self)

            @property
            def _premake_configuration(self):
                return str(self.settings.build_type) + "Sanitizer" if self.options.sanitizer else ""

            def generate(self):
                deps = PremakeDeps(self)
                deps.configuration = self._premake_configuration
                deps.generate()
                tc = PremakeToolchain(self)
                tc.generate()

            def build(self):
                premake = Premake(self)
                premake.configure()
                # premake.build(workspace="Consumer")
                premake.build(workspace="Consumer", configuration=self._premake_configuration)
    """)

    c.save({"src/main.cpp": gen_function_cpp(name="main"),
            "premake5.lua": gen_premake5(
                workspace="Consumer",
                projects=[
                    {
                        "name": "consumer",
                        "files": ["src/main.cpp"],
                        "kind": "ConsoleApp",
                    }
                ],
                configurations=["Debug", "Release", "DebugSanitizer", "ReleaseSanitizer"],
            ),
            "conanfile.py": conanfile,
    })
    # Test it builds successfully with the custom configuration
    c.run("build . -o sanitizer=True")



from conan.internal.model.conf import Conf


def test_copy_conaninfo_conf():
    conf = Conf()

    conf.define("core:non_interactive", True)
    conf.define("tools.cmake.cmaketoolchain:generator", True)
    conf.define("tools.deployer:symlinks", True)
    conf.define("user.myconf:cmake-test", True)

    pattern = [".*"]
    conf.define("tools.info.package_id:confs", pattern)
    result = conf.copy_conaninfo_conf().dumps()
    assert "tools.info.package_id:confs=%s" % pattern in result
    assert "core:non_interactive" in result
    assert "tools.cmake.cmaketoolchain:generator=True" in result
    assert "tools.deployer:symlinks" in result
    assert "user.myconf:cmake-test" in result

    pattern = [r"tools\..*"]
    conf.define("tools.info.package_id:confs", pattern)
    result = conf.copy_conaninfo_conf().dumps()
    assert "tools.info.package_id:confs=%s" % pattern in result
    assert "core:non_interactive" not in result
    assert "tools.cmake.cmaketoolchain:generator=True" in result
    assert "tools.deployer:symlinks" in result
    assert "user.myconf:cmake-test" not in result

    pattern = [".*cmake"]
    conf.define("tools.info.package_id:confs", pattern)
    result = conf.copy_conaninfo_conf().dumps()
    assert "tools.info.package_id:confs=%s" % pattern not in result
    assert "core:non_interactive" not in result
    assert "tools.cmake.cmaketoolchain:generator=True" in result
    assert "tools.deployer:symlinks" not in result
    assert "user.myconf:cmake-test" in result

    pattern = ["(tools.deploy|core)"]
    conf.define("tools.info.package_id:confs", pattern)
    result = conf.copy_conaninfo_conf().dumps()
    assert "tools.info.package_id:confs=%s" % pattern not in result
    assert "core:non_interactive" in result
    assert "tools.cmake.cmaketoolchain:generator=True" not in result
    assert "tools.deployer:symlinks" in result
    assert "user.myconf:cmake-test" not in result




import textwrap

import pytest

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient


@pytest.mark.parametrize("requires", ["requires", "tool_requires"])
def test_conanfile_txt_deps_revisions(requires):
    """
    conanfile.txt locking it dependencies (with revisions)
    """
    client = TestClient(light=True)
    client.save({"pkg/conanfile.py": GenConanfile().with_package_id("self.output.info('REV1!!!!')"),
                 "consumer/conanfile.txt": f"[{requires}]\npkg/0.1@user/testing"})
    client.run("create pkg --name=pkg --version=0.1 --user=user --channel=testing")
    assert "REV1!!!" in client.out
    client.run("lock create consumer/conanfile.txt  --lockfile-out=consumer.lock")
    assert "pkg/0.1@user/testing#" in client.out

    client.save({"pkg/conanfile.py": GenConanfile().with_package_id("self.output.info('REV2!!!!')")})
    client.run("create pkg --name=pkg --version=0.1 --user=user --channel=testing")
    assert "REV2!!!" in client.out

    client.run("install consumer/conanfile.txt --lockfile=consumer.lock")
    assert "REV1!!!" in client.out
    assert "REV2!!!" not in client.out
    client.run("install consumer/conanfile.txt")
    assert "REV2!!!" in client.out
    assert "REV1!!!" not in client.out


@pytest.mark.parametrize("requires", ["requires", "tool_requires"])
@pytest.mark.parametrize("req_version", ["0.1", "[>=0.0]"])
def test_conanfile_txt_deps_revisions_transitive(requires, req_version):
    """
    conanfile.txt locking it dependencies and its transitive dependencies (with revisions)
    """
    client = TestClient(light=True)
    client.save({"dep/conanfile.py": GenConanfile().with_package_id("self.output.info('REV1!!!!')"),
                 "pkg/conanfile.py": GenConanfile().with_requires(f"dep/{req_version}@user/testing"),
                 "consumer/conanfile.txt": f"[{requires}]\npkg/{req_version}@user/testing"})
    client.run("create dep --name=dep --version=0.1 --user=user --channel=testing")
    assert "REV1!!!" in client.out
    client.run("create pkg --name=pkg --version=0.1 --user=user --channel=testing")

    client.run("lock create consumer/conanfile.txt  --lockfile-out=consumer.lock")
    assert "dep/0.1@user/testing#" in client.out
    assert "pkg/0.1@user/testing#" in client.out

    client.save({"dep/conanfile.py": GenConanfile().with_package_id("self.output.info('REV2!!!!')")})
    client.run("create dep --name=dep --version=0.1 --user=user --channel=testing")
    assert "REV2!!!" in client.out

    client.run("install consumer/conanfile.txt --lockfile=consumer.lock")
    assert "REV1!!!" in client.out
    assert "REV2!!!" not in client.out
    client.run("list dep/0.1@user/testing#*")
    client.run("install consumer/conanfile.txt")
    assert "REV2!!!" in client.out
    assert "REV1!!!" not in client.out


@pytest.mark.parametrize("requires", ["requires", "tool_requires"])
def test_conanfile_txt_strict_revisions(requires):
    """
    conanfile.txt locking it dependencies (with version ranges)
    """
    client = TestClient(light=True)
    client.save({"pkg/conanfile.py": GenConanfile().with_package_id("self.output.info('REV1!!!!')"),
                 "consumer/conanfile.txt": f"[{requires}]\npkg/0.1@user/testing"})
    client.run("create pkg --name=pkg --version=0.1 --user=user --channel=testing")
    client.run("lock create consumer/conanfile.txt")
    assert "pkg/0.1@user/testing#" in client.out

    client.save({"pkg/conanfile.py": GenConanfile().with_package_id("self.output.info('REV2!!!!')")})
    client.run("create pkg --name=pkg --version=0.1 --user=user --channel=testing")
    rrev = client.exported_recipe_revision()

    # Not strict mode works
    client.save({"consumer/conanfile.txt": f"[{requires}]\npkg/0.1@user/testing#{rrev}"})

    client.run("install consumer/conanfile.txt", assert_error=True)
    assert f"Requirement 'pkg/0.1@user/testing#{rrev}' not in lockfile" in client.out


@pytest.mark.parametrize("requires", ["requires", "tool_requires"])
def test_conditional_os(requires):
    """
    conanfile.txt can lock conditional dependencies (conditional on OS for example),
    with consecutive calls to "conan lock create", augmenting the lockfile
    """
    client = TestClient(light=True)

    pkg_conanfile = textwrap.dedent(f"""
        from conan import ConanFile
        class Pkg(ConanFile):
            settings = "os"
            def requirements(self):
                if self.settings.os == "Windows":
                    self.requires("win/0.1")
                else:
                    self.requires("nix/0.1")
        """)

    client.save({"dep/conanfile.py": GenConanfile().with_package_id("self.output.info('REV1!!!!')"),
                 "pkg/conanfile.py": pkg_conanfile,
                 "consumer/conanfile.txt": f"[{requires}]\npkg/0.1"})
    client.run("create dep --name=win --version=0.1")
    client.run("create dep --name=nix --version=0.1")

    client.run("create pkg --name=pkg --version=0.1 -s os=Windows")
    client.run("create pkg --name=pkg --version=0.1 -s os=Linux")

    client.run("lock create consumer/conanfile.txt  --lockfile-out=consumer.lock -s os=Windows"
               " -s:b os=Windows")
    assert "win/0.1#" in client.out
    assert "pkg/0.1#" in client.out
    client.run("lock create consumer/conanfile.txt  --lockfile=consumer.lock "
               "--lockfile-out=consumer.lock -s os=Linux -s:b os=Linux")
    assert "nix/0.1#" in client.out
    assert "pkg/0.1#" in client.out

    # New dependencies will not be used if using the lockfile
    client.save({"dep/conanfile.py": GenConanfile().with_package_id("self.output.info('REV2!!!!')")})
    client.run("create dep --name=win --version=0.1")
    client.run("create dep --name=nix --version=0.1")
    client.run("create pkg --name=pkg --version=0.1 -s os=Windows")
    client.run("create pkg --name=pkg --version=0.1 -s os=Linux")

    client.run("install consumer --lockfile=consumer.lock -s os=Windows -s:b os=Windows")
    assert "REV1!!!" in client.out
    assert "REV2!!!" not in client.out
    assert "nix/0.1" not in client.out
    client.run("install consumer -s os=Windows -s:b os=Windows")
    assert "REV2!!!" in client.out
    assert "REV1!!!" not in client.out
    assert "nix/0.1" not in client.out

    client.run("install consumer --lockfile=consumer.lock -s os=Linux -s:b os=Linux")
    assert "REV1!!!" in client.out
    assert "REV2!!!" not in client.out
    assert "win/0.1" not in client.out
    client.run("install consumer -s os=Linux -s:b os=Linux")
    assert "REV2!!!" in client.out
    assert "REV1!!!" not in client.out
    assert "win/0.1" not in client.out


@pytest.mark.parametrize("requires", ["requires", "tool_requires"])
def test_conditional_same_package_revisions(requires):
    # What happens when a conditional requires different versions of the same package?
    client = TestClient(light=True)

    pkg_conanfile = textwrap.dedent("""
        from conan import ConanFile
        class Pkg(ConanFile):
            settings = "os"
            def requirements(self):
                if self.settings.os == "Windows":
                    self.requires("dep/0.1#{}")
                else:
                    self.requires("dep/0.1#{}")
        """)
    client.save({"dep1/conanfile.py": GenConanfile().with_package_id("self.output.info('REV1!!!!')"),
                 "dep2/conanfile.py": GenConanfile().with_package_id("self.output.info('REV2!!!!')"),
                 "pkg/conanfile.py": pkg_conanfile,
                 "consumer/conanfile.txt": f"[{requires}]\npkg/0.1"})
    client.run("create dep1 --name=dep --version=0.1")
    rrev1 = client.exported_recipe_revision()
    client.run("create dep2 --name=dep --version=0.1")
    rrev2 = client.exported_recipe_revision()
    client.save({"pkg/conanfile.py": pkg_conanfile.format(rrev1, rrev2)})

    client.run("create pkg --name=pkg --version=0.1 -s os=Windows")
    client.run("create pkg --name=pkg --version=0.1 -s os=Linux")

    client.run("lock create consumer/conanfile.txt  --lockfile-out=conan.lock -s os=Windows"
               " -s:b os=Windows")
    assert "REV1!!!" in client.out
    assert "REV2!!!" not in client.out
    client.run("lock create consumer/conanfile.txt  --lockfile=conan.lock "
               "--lockfile-out=conan.lock -s os=Linux -s:b os=Linux")
    assert "REV2!!!" in client.out
    assert "REV1!!!" not in client.out

    client.run("install consumer --lockfile=conan.lock --lockfile-out=win.lock -s os=Windows"
               " -s:b os=Windows")
    assert "REV1!!!" in client.out
    assert "REV2!!!" not in client.out

    client.run("install consumer --lockfile=conan.lock --lockfile-out=linux.lock -s os=Linux"
               " -s:b os=Linux")
    assert "REV2!!!" in client.out
    assert "REV1!!!" not in client.out



import os
import textwrap

from conan.test.utils.tools import TestClient, GenConanfile


class TestBuildRequiresFromProfile:
    profile_host = textwrap.dedent("""
        [settings]
        os=Windows
        arch=x86_64
         compiler=msvc
        compiler.version=192
        compiler.runtime=dynamic

        [tool_requires]
        br2/version
    """)

    profile_build = textwrap.dedent("""
        [settings]
        os=Macos
        arch=x86_64
        compiler=apple-clang
        compiler.version=11.0
        compiler.libcxx=libc++
        build_type=Release

        [tool_requires]
        br3/version
    """)

    library_conanfile = textwrap.dedent("""
        from conan import ConanFile

        class Recipe(ConanFile):
            name = "library"
            version = "version"

            build_requires = "br1/version"
    """)

    def test_br_from_profile_host_and_profile_build(self):
        t = TestClient()
        t.save({'profile_host': self.profile_host,
                'profile_build': self.profile_build,
                'library.py': self.library_conanfile,
                'br1.py': GenConanfile(),
                'br2.py': GenConanfile(),
                'br3.py': GenConanfile()})
        t.run("export br1.py --name=br1 --version=version")
        t.run("export br2.py --name=br2 --version=version")
        t.run("export br3.py --name=br3 --version=version")
        t.run("create library.py --profile:host=profile_host --profile:build=profile_build "
              "--build='*'")


class TestBuildRequiresContextHostFromProfile:
    toolchain = textwrap.dedent("""
        from conan import ConanFile

        class Recipe(ConanFile):
            name = "mytoolchain"
            version = "1.0"
            settings = "os"

            def package_info(self):
                self.output.info("PackageInfo OS=%s" % self.settings.os)
        """)
    gtest = textwrap.dedent("""
        from conan import ConanFile
        import os

        class Recipe(ConanFile):
            name = "gtest"
            version = "1.0"
            settings = "os"

            def build(self):
                self.output.info("Build OS=%s" % self.settings.os)
            def package_info(self):
                self.output.info("PackageInfo OS=%s" % self.settings.os)
        """)
    library_conanfile = textwrap.dedent("""
         from conan import ConanFile
         import os

         class Recipe(ConanFile):
             name = "library"
             version = "version"
             settings = "os"

             def build_requirements(self):
                self.test_requires("gtest/1.0")

             def build(self):
                self.output.info("Build OS=%s" % self.settings.os)
         """)
    profile_host = textwrap.dedent("""
        [settings]
        os = Linux
        [tool_requires]
        mytoolchain/1.0
        """)
    profile_build = textwrap.dedent("""
        [settings]
        os = Windows
        """)

    def test_br_from_profile_host_and_profile_build(self):
        t = TestClient()
        t.save({'profile_host': self.profile_host,
                'profile_build': self.profile_build,
                'library.py': self.library_conanfile,
                'mytoolchain.py': self.toolchain,
                "gtest.py": self.gtest})
        t.run("create mytoolchain.py -pr:h=profile_host -pr:b=profile_build --build-require")

        t.run("create gtest.py -pr=profile_host -pr:b=profile_build")
        assert "mytoolchain/1.0: PackageInfo OS=Windows" in t.out
        assert "gtest/1.0: PackageInfo OS=Linux" in t.out

        t.run("create gtest.py -pr=profile_host -pr:b=profile_build --build=*")
        assert "mytoolchain/1.0: PackageInfo OS=Windows" in t.out
        assert "gtest/1.0: Build OS=Linux" in t.out
        assert "gtest/1.0: PackageInfo OS=Linux" in t.out

        t.run("create library.py -pr:h=profile_host -pr:b=profile_build")
        assert "gtest/1.0: PackageInfo OS=Linux" in t.out
        assert "library/version: Build OS=Linux" in t.out
        assert "mytoolchain/1.0: PackageInfo OS=Windows" in t.out

        t.run("create library.py -pr:h=profile_host -pr:b=profile_build --build=*")
        assert "gtest/1.0: Build OS=Linux" in t.out
        assert "gtest/1.0: PackageInfo OS=Linux" in t.out
        assert "library/version: Build OS=Linux" in t.out
        assert "mytoolchain/1.0: PackageInfo OS=Windows" in t.out


class TestBuildRequiresBothContexts:
    toolchain_creator = textwrap.dedent("""
        from conan import ConanFile

        class Recipe(ConanFile):
            name = "creator"
            version = "1.0"
            settings = "os"

            def package_info(self):
                self.output.info("PackageInfo OS=%s" % self.settings.os)
        """)
    toolchain = textwrap.dedent("""
        from conan import ConanFile
        import os

        class Recipe(ConanFile):
            name = "mytoolchain"
            version = "1.0"
            settings = "os"
            def build(self):
                self.output.info("Build OS=%s" % self.settings.os)

            def package_info(self):
                self.output.info("PackageInfo OS=%s" % self.settings.os)
        """)
    gtest = textwrap.dedent("""
        from conan import ConanFile
        import os

        class Recipe(ConanFile):
            name = "gtest"
            version = "1.0"
            settings = "os"

            def build(self):
                self.output.info("Build OS=%s" % self.settings.os)
            def package_info(self):
                self.output.info("PackageInfo OS=%s" % self.settings.os)
        """)
    library_conanfile = textwrap.dedent("""
         from conan import ConanFile
         import os

         class Recipe(ConanFile):
             name = "library"
             version = "version"
             settings = "os"

             def build_requirements(self):
                self.test_requires("gtest/1.0")

             def build(self):
                self.output.info("Build OS=%s" % self.settings.os)
         """)
    profile_host = textwrap.dedent("""
        [settings]
        os = Linux
        [tool_requires]
        mytoolchain/1.0
        """)
    profile_build = textwrap.dedent("""
        [settings]
        os = Windows
        [tool_requires]
        mytoolchain*:creator/1.0
        """)

    def test_build_requires_both_contexts(self):
        t = TestClient()
        t.save({'profile_host': self.profile_host,
                'profile_build': self.profile_build,
                'library.py': self.library_conanfile,
                'creator.py': self.toolchain_creator,
                'mytoolchain.py': self.toolchain,
                "gtest.py": self.gtest})
        t.run("create creator.py -pr=profile_build")
        t.run("create mytoolchain.py -pr:h=profile_host -pr:b=profile_build --build-require")
        assert "creator/1.0: PackageInfo OS=Windows" in t.out
        assert "mytoolchain/1.0: Build OS=Windows" in t.out

        # new way, the toolchain can now run in Windows, but gtest in Linux
        t.run("create gtest.py --profile=profile_host --profile:build=profile_build")
        assert "creator/1.0: PackageInfo" not in t.out  # Creator is skipped now, not needed
        assert "gtest/1.0: PackageInfo OS=Linux" in t.out

        t.run("create gtest.py --profile=profile_host --profile:build=profile_build --build=*")
        assert "creator/1.0: PackageInfo OS=Windows" in t.out
        assert "gtest/1.0: PackageInfo OS=Linux" in t.out

        # Declaring the build_requires in the recipe works, it is just the profile that is
        # not transitive
        toolchain = textwrap.dedent("""
            from conan import ConanFile
            import os

            class Recipe(ConanFile):
                name = "mytoolchain"
                version = "1.0"
                settings = "os"
                build_requires = "creator/1.0"
                def build(self):
                    self.output.info("Build OS=%s" % self.settings.os)

                def package_info(self):
                    self.output.info("PackageInfo OS=%s" % self.settings.os)
            """)
        t.save({'mytoolchain.py': toolchain})
        t.run("create mytoolchain.py --profile:host=profile_build -pr:b=profile_build")
        assert "creator/1.0: PackageInfo OS=Windows" in t.out
        assert "mytoolchain/1.0: Build OS=Windows" in t.out

        t.run("create gtest.py --profile=profile_host --profile:build=profile_build --build=*")
        assert "creator/1.0: PackageInfo OS=Windows" in t.out
        assert "mytoolchain/1.0: Build OS=Windows" in t.out
        assert "gtest/1.0: Build OS=Linux" in t.out

        t.run("create library.py -pr:h=profile_host --profile:build=profile_build --build=*")
        assert "creator/1.0: PackageInfo OS=Windows" in t.out
        assert "mytoolchain/1.0: Build OS=Windows" in t.out
        assert "gtest/1.0: Build OS=Linux" in t.out
        assert "gtest/1.0: PackageInfo OS=Linux" in t.out
        assert "library/version: Build OS=Linux" in t.out


def test_consumer_get_profile_tool_requires():
    # [tool_requires] defined in a profile will result in applying them in the
    # consumer, both if using a conanfile.txt of using a --requires
    t = TestClient(light=True)
    tool = textwrap.dedent("""
        from conan import ConanFile
        class Pkg(ConanFile):
            name = "tool"
            version = "1.0"
            def package_info(self):
                self.buildenv_info.define("MYVAR", "MYVALUE")
        """)
    t.save({'profile': "[tool_requires]\ntool/1.0",
            'dep.py': GenConanfile("dep", "1.0"),
            'tool.py': tool,
            'conanfile.txt': ""})
    t.run("create tool.py")
    t.run("create dep.py")
    t.run("install . --profile ./profile")
    env = t.load("conanbuildenv.sh")
    assert 'export MYVAR="MYVALUE"' in env
    os.remove(os.path.join(t.current_folder, "conanbuildenv.sh"))
    # also when using --requires
    t.run("install --requires=dep/1.0 --profile ./profile")
    env = t.load("conanbuildenv.sh")
    assert 'export MYVAR="MYVALUE"' in env



import json
import textwrap

import pytest

from conan.api.model import RecipeReference
from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient


@pytest.mark.parametrize("require, pattern, alternative, pkg", [
    # PATTERN VERSIONS
    # override all dependencies to "dep" to a specific version,user and channel)
    # TODO: This is a version override, is this really wanted?
    ("dep/1.3", "dep/*", "dep/1.1", "dep/1.1"),
    ("dep/[>=1.0 <2]", "dep/*", "dep/1.1", "dep/1.1"),
    # override all dependencies to "dep" to the same version with other user, remove channel)
    ("dep/1.3", "dep/*", "dep/*@system", "dep/1.3@system"),
    ("dep/[>=1.0 <2]", "dep/*", "dep/*@system", "dep/1.1@system"),
    # override all dependencies to "dep" to the same version with other user, same channel)
    ("dep/1.3@comp/stable", "dep/*@*/*", "dep/*@system/*", "dep/1.3@system/stable"),
    ("dep/[>=1.0 <2]@comp/stable", "dep/*@*/*", "dep/*@system/*", "dep/1.1@system/stable"),
    # EXACT VERSIONS
    # replace exact dependency version for one in the system
    ("dep/1.1", "dep/1.1", "dep/1.1@system", "dep/1.1@system"),
    ("dep/[>=1.0 <2]", "dep/1.1", "dep/1.1@system", "dep/1.1@system"),
    ("dep/[>=1.0 <2]@comp", "dep/1.1@*", "dep/1.1@*/stable", "dep/1.1@comp/stable"),
    ("dep/1.1@comp", "dep/1.1@*", "dep/1.1@*/stable", "dep/1.1@comp/stable"),
    # PACKAGE ALTERNATIVES (zlib->zlibng)
    ("dep/1.0", "dep/*", "depng/*", "depng/1.0"),
    ("dep/[>=1.0 <2]", "dep/*", "depng/*", "depng/1.1"),
    ("dep/[>=1.0 <2]", "dep/1.1", "depng/1.2", "depng/1.2"),
    # NON MATCHING
    ("dep/1.3", "dep/1.1", "dep/1.1@system", "dep/1.3"),
    ("dep/1.3", "dep/*@comp", "dep/*@system", "dep/1.3"),
    ("dep/[>=1.0 <2]", "dep/2.1", "dep/2.1@system", "dep/1.1"),
    # PATTERN - PATTERN REPLACE
    ("dep/[>=1.3 <2]", "dep/*", "dep/[>=1.0 <1.9]", "dep/1.1"),
    # DIRECT REPLACE OF PINNED VERSIONS
    ("dep/1.3", "dep/1.3", "dep/1.5", "dep/1.5"),
])
@pytest.mark.parametrize("tool_require", [False, True])
class TestReplaceRequires:
    def test_alternative(self, tool_require, require, pattern, alternative, pkg):
        c = TestClient(light=True)
        conanfile = GenConanfile().with_tool_requires(require) if tool_require else \
            GenConanfile().with_requires(require)
        profile_tag = "replace_requires" if not tool_require else "replace_tool_requires"
        c.save({"dep/conanfile.py": GenConanfile(),
                "pkg/conanfile.py": conanfile,
                "profile": f"[{profile_tag}]\n{pattern}: {alternative}"})
        ref = RecipeReference.loads(pkg)
        user = f"--user={ref.user}" if ref.user else ""
        channel = f"--channel={ref.channel}" if ref.channel else ""
        c.run(f"create dep --name={ref.name} --version={ref.version} {user} {channel}")
        rrev = c.exported_recipe_revision()
        c.run("profile show -pr=profile")
        assert profile_tag in c.out
        c.run("install pkg -pr=profile")
        assert profile_tag in c.out
        c.assert_listed_require({f"{pkg}#{rrev}": "Cache"}, build=tool_require)

        # Check lockfile
        c.run("lock create pkg -pr=profile")
        lock = c.load("pkg/conan.lock")
        assert f"{pkg}#{rrev}" in lock

        # c.run("create dep2 --version=1.2")
        # with lockfile
        c.run("install pkg -pr=profile")
        c.assert_listed_require({f"{pkg}#{rrev}": "Cache"}, build=tool_require)

    def test_diamond(self, tool_require, require, pattern, alternative, pkg):
        c = TestClient(light=True)
        conanfile = GenConanfile().with_tool_requires(require) if tool_require else \
            GenConanfile().with_requires(require)
        profile_tag = "replace_requires" if not tool_require else "replace_tool_requires"

        c.save({"dep/conanfile.py": GenConanfile(),
                "libb/conanfile.py": conanfile,
                "libc/conanfile.py": conanfile,
                "app/conanfile.py": GenConanfile().with_requires("libb/0.1", "libc/0.1"),
                "profile": f"[{profile_tag}]\n{pattern}: {alternative}"})
        ref = RecipeReference.loads(pkg)
        user = f"--user={ref.user}" if ref.user else ""
        channel = f"--channel={ref.channel}" if ref.channel else ""
        c.run(f"create dep --name={ref.name} --version={ref.version} {user} {channel}")
        rrev = c.exported_recipe_revision()

        c.run("export libb --name=libb --version=0.1")
        c.run("export libc --name=libc --version=0.1")

        c.run("install app -pr=profile", assert_error=True)
        assert "ERROR: Missing binary: libb/0.1" in c.out
        assert "ERROR: Missing binary: libc/0.1" in c.out

        c.run("install app -pr=profile --build=missing")
        c.assert_listed_require({f"{pkg}#{rrev}": "Cache"}, build=tool_require)

        # Check lockfile
        c.run("lock create app -pr=profile")
        lock = c.load("app/conan.lock")
        assert f"{pkg}#{rrev}" in lock

        # with lockfile
        c.run("install app -pr=profile")
        c.assert_listed_require({f"{pkg}#{rrev}": "Cache"}, build=tool_require)


@pytest.mark.parametrize("pattern, replace", [
    ("pkg", "pkg/0.1"),
    ("pkg/*", "pkg"),
    ("pkg/*:pid1", "pkg/0.1"),
    ("pkg/*:pid1", "pkg/0.1:pid2"),
    ("pkg/*", "pkg/0.1:pid2"),
    (":", ""),
    ("pkg/version:pid", ""),
    ("pkg/version:pid", ":")
])
def test_replace_requires_errors(pattern, replace):
    c = TestClient(light=True)
    c.save({"pkg/conanfile.py": GenConanfile("pkg", "0.1"),
            "app/conanfile.py": GenConanfile().with_requires("pkg/0.2"),
            "profile": f"[replace_requires]\n{pattern}: {replace}"})
    c.run("create pkg")
    c.run("install app -pr=profile", assert_error=True)
    assert "ERROR: Error reading 'profile' profile: Error in [replace_xxx]" in c.out


def test_replace_requires_invalid_requires_errors():
    """
    replacing for something incorrect not existing is not an error per-se, it is valid that
    a recipe requires("pkg/2.*"), and then it will fail because such package doesn't exist
    """
    c = TestClient(light=True)
    c.save({"app/conanfile.py": GenConanfile().with_requires("pkg/0.2"),
            "profile": f"[replace_requires]\npkg/0.2: pkg/2.*"})
    c.run("install app -pr=profile", assert_error=True)
    assert "pkg/0.2: pkg/2.*" in c.out  # The replacement happens
    assert "ERROR: Package 'pkg/2.*' not resolved" in c.out


def test_replace_requires_json_format():
    c = TestClient(light=True)
    c.save({"pkg/conanfile.py": GenConanfile("pkg", "0.2"),
            "app/conanfile.py": GenConanfile().with_requires("pkg/0.1"),
            "profile": f"[replace_requires]\npkg/0.1: pkg/0.2"})
    c.run("create pkg")
    c.run("install app -pr=profile --format=json")
    assert "pkg/0.1: pkg/0.2" in c.out  # The replacement happens
    graph = json.loads(c.stdout)
    assert graph["graph"]["replaced_requires"] == {"pkg/0.1": "pkg/0.2"}
    assert graph["graph"]["nodes"]["0"]["dependencies"]["1"]["ref"] == "pkg/0.2"
    assert graph["graph"]["nodes"]["0"]["dependencies"]["1"]["require"] == "pkg/0.1"


def test_replace_requires_test_requires():
    c = TestClient(light=True)
    c.save({"gtest/conanfile.py": GenConanfile("gtest", "0.2"),
            "app/conanfile.py": GenConanfile().with_test_requires("gtest/0.1"),
            "profile": f"[replace_requires]\ngtest/0.1: gtest/0.2"})
    c.run("create gtest")
    c.run("install app -pr=profile")
    assert "gtest/0.1: gtest/0.2" in c.out  # The replacement happens


# We test even replacing by itself, not great, but shouldn't crash
@pytest.mark.parametrize("name, version", [("zlib", "0.1"), ("zlib", "0.2"), ("zlib-ng", "0.1")])
def test_replace_requires_consumer_references(name, version):
    c = TestClient()
    # IMPORTANT: The replacement package must be target-compatible
    dep = textwrap.dedent(f"""
        from conan import ConanFile
        class ZlibNG(ConanFile):
            name = "{name}"
            version = "{version}"
            def package_info(self):
                self.cpp_info.set_property("cmake_file_name", "ZLIB")
                self.cpp_info.set_property("cmake_target_name", "ZLIB::ZLIB")
        """)
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        class App(ConanFile):
            name = "app"
            version = "0.1"
            settings = "build_type"
            requires = "zlib/0.1"
            generators = "CMakeDeps"

            def generate(self):
                self.output.info(f"DEP ZLIB generate: {self.dependencies['zlib'].ref.name}!")
            def build(self):
                self.output.info(f"DEP ZLIB build: {self.dependencies['zlib'].ref.name}!")
            def package_info(self):
                self.output.info(f"DEP ZLIB package_info: {self.dependencies['zlib'].ref.name}!")
                self.cpp_info.requires = ["zlib::zlib"]
        """)
    c.save({"dep/conanfile.py": dep,
            "app/conanfile.py": conanfile,
            "profile": f"[replace_requires]\nzlib/0.1: {name}/{version}"})
    c.run("create dep")
    c.run("build app -pr=profile")
    assert f"zlib/0.1: {name}/{version}" in c.out
    assert f"conanfile.py (app/0.1): DEP ZLIB generate: {name}!" in c.out
    assert f"conanfile.py (app/0.1): DEP ZLIB build: {name}!" in c.out
    # Check generated CMake code. If the targets are NOT compatible, then the replacement
    # Cannot happen
    assert "find_package(ZLIB)" in c.out
    assert "target_link_libraries(... ZLIB::ZLIB)" in c.out
    cmake = c.load("app/ZLIBTargets.cmake")
    assert "add_library(ZLIB::ZLIB INTERFACE IMPORTED)" in cmake
    c.run("create app -pr=profile")
    assert f"zlib/0.1: {name}/{version}" in c.out
    assert f"app/0.1: DEP ZLIB generate: {name}!" in c.out
    assert f"app/0.1: DEP ZLIB build: {name}!" in c.out
    if name == "zlib-ng":
        # CMakeDeps can not be used to consume replaced requires for different packages
        # only CMakeConfigDeps has this capability
        c.run("install --requires=app/0.1 -pr=profile -g CMakeDeps", assert_error=True)


def test_replace_requires_consumer_references_error_multiple():
    # https://github.com/conan-io/conan/issues/17407
    c = TestClient()
    # IMPORTANT: The replacement package must be target-compatible
    zlib = textwrap.dedent("""
        from conan import ConanFile
        class ZlibNG(ConanFile):
            name = "zlib"
            version = "0.2"
            def package_info(self):
                self.cpp_info.set_property("cmake_file_name", "ZLIB")
                self.cpp_info.set_property("cmake_target_name", "ZLIB::ZLIB")
        """)
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        class App(ConanFile):
            name = "app"
            version = "0.1"
            settings = "build_type"
            requires = "zlib/0.1", "bzip2/0.1"
            generators = "CMakeDeps"

            def generate(self):
                self.output.info(f"DEP ZLIB generate: {self.dependencies['zlib'].ref.name}!")
                self.output.info(f"DEP BZIP2 generate: {self.dependencies['bzip2'].ref.name}!")
            def build(self):
                self.output.info(f"DEP ZLIB build: {self.dependencies['zlib'].ref.name}!")
                self.output.info(f"DEP BZIP2 build: {self.dependencies['bzip2'].ref.name}!")
            def package_info(self):
                self.output.info(f"DEP ZLIB package_info: {self.dependencies['zlib'].ref.name}!")
                self.cpp_info.requires = ["zlib::zlib", "bzip2::bzip2"]
        """)
    c.save({"zlib/conanfile.py": zlib,
            "app/conanfile.py": conanfile,
            "profile": "[replace_requires]\nzlib/0.1: zlib/0.2\nbzip2/0.1: zlib/0.2"})
    c.run("create zlib")
    c.run("build app -pr=profile")
    assert "zlib/0.1: zlib/0.2" in c.out
    assert "conanfile.py (app/0.1): DEP ZLIB generate: zlib!" in c.out
    assert "conanfile.py (app/0.1): DEP ZLIB build: zlib!" in c.out
    assert "conanfile.py (app/0.1): DEP BZIP2 generate: zlib!" in c.out
    assert "conanfile.py (app/0.1): DEP BZIP2 build: zlib!" in c.out
    # Check generated CMake code. If the targets are NOT compatible, then the replacement
    # Cannot happen
    assert "find_package(ZLIB)" in c.out
    assert "target_link_libraries(... ZLIB::ZLIB ZLIB::ZLIB)" in c.out
    cmake = c.load("app/ZLIBTargets.cmake")
    assert "add_library(ZLIB::ZLIB INTERFACE IMPORTED)" in cmake
    c.run("create app -pr=profile")
    assert "zlib/0.1: zlib/0.2" in c.out
    assert "app/0.1: DEP ZLIB generate: zlib!" in c.out
    assert "app/0.1: DEP ZLIB build: zlib!" in c.out


def test_replace_requires_consumer_components_options():
    c = TestClient()
    # IMPORTANT: The replacement package must be target-compatible
    zlib_ng = textwrap.dedent("""
        from conan import ConanFile
        class ZlibNG(ConanFile):
            name = "zlib-ng"
            version = "0.1"
            options = {"compat": [False, True]}
            default_options = {"compat": False}
            def package_info(self):
                self.cpp_info.set_property("cmake_file_name", "ZLIB")
                self.cpp_info.set_property("cmake_target_name", "ZLIB::ZLIB")
                if self.options.compat:
                    self.cpp_info.components["myzlib"].set_property("cmake_target_name",
                                                                    "ZLIB::zmylib")
        """)
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        class App(ConanFile):
            name = "app"
            version = "0.1"
            settings = "build_type"
            requires = "zlib/0.1"
            generators = "CMakeDeps"

            def generate(self):
                self.output.info(f"DEP ZLIB generate: {self.dependencies['zlib'].ref.name}!")
            def build(self):
                self.output.info(f"DEP ZLIB build: {self.dependencies['zlib'].ref.name}!")
            def package_info(self):
                self.output.info(f"zlib in deps?: {'zlib' in self.dependencies}")
                self.output.info(f"zlib-ng in deps?: {'zlib-ng' in self.dependencies}")
                self.output.info(f"DEP ZLIB package_info: {self.dependencies['zlib'].ref.name}!")
                self.cpp_info.requires = ["zlib::myzlib"]
        """)
    profile = textwrap.dedent("""
        [options]
        zlib-ng/*:compat=True

        [replace_requires]
        zlib/0.1: zlib-ng/0.1
        """)
    c.save({"zlibng/conanfile.py": zlib_ng,
            "app/conanfile.py": conanfile,
            "profile": profile})

    c.run("create zlibng -o *:compat=True")
    c.run("build app -pr=profile")
    assert "zlib/0.1: zlib-ng/0.1" in c.out
    assert "conanfile.py (app/0.1): DEP ZLIB generate: zlib-ng!" in c.out
    assert "conanfile.py (app/0.1): DEP ZLIB build: zlib-ng!" in c.out
    # Check generated CMake code. If the targets are NOT compatible, then the replacement
    # Cannot happen
    assert "find_package(ZLIB)" in c.out
    assert "target_link_libraries(... ZLIB::ZLIB)" in c.out
    cmake = c.load("app/ZLIBTargets.cmake")
    assert "add_library(ZLIB::ZLIB INTERFACE IMPORTED)" in cmake
    cmake = c.load("app/ZLIB-Target-none.cmake")
    assert "set_property(TARGET ZLIB::ZLIB APPEND PROPERTY INTERFACE_LINK_LIBRARIES ZLIB::zmylib)" \
           in cmake

    c.run("create app -pr=profile")
    assert "zlib/0.1: zlib-ng/0.1" in c.out
    assert "app/0.1: DEP ZLIB generate: zlib-ng!" in c.out
    assert "app/0.1: DEP ZLIB build: zlib-ng!" in c.out
    assert "find_package(ZLIB)" in c.out
    assert "target_link_libraries(... ZLIB::ZLIB)" in c.out
    assert "zlib in deps?: True" in c.out
    assert "zlib-ng in deps?: False" in c.out


def test_replace_requires_multiple():
    conanfile = textwrap.dedent("""
        from conan import ConanFile

        class EpoxyConan(ConanFile):
            name = "libepoxy"
            version = "0.1"

            def requirements(self):
                self.requires("opengl/system")
                self.requires("egl/system")

            def generate(self):
                for r, d in self.dependencies.items():
                    self.output.info(f"DEP: {r.ref.name}: {d.ref.name}")

            def package_info(self):
                self.cpp_info.requires.append("opengl::opengl")
                self.cpp_info.requires.append("egl::egl")
        """)
    profile = textwrap.dedent("""
        [replace_requires]
        opengl/system: libgl/1.0
        egl/system: libgl/1.0
        """)
    c = TestClient()
    c.save({"dep/conanfile.py": GenConanfile(),
            "app/conanfile.py": conanfile,
            "profile": profile})
    c.run("create dep --name=libgl --version=1.0")
    c.run("create app -pr=profile")
    # There are actually 2 dependencies, pointing to the same node
    assert "libepoxy/0.1: DEP: opengl: libgl" in c.out
    assert "libepoxy/0.1: DEP: egl: libgl" in c.out


class TestReplaceRequiresTransitiveGenerators:
    """ Generators are incorrectly managing replace_requires
    # https://github.com/conan-io/conan/issues/17557
    """

    @pytest.mark.parametrize("diamond", [True, False])
    def test_no_components(self, diamond):
        c = TestClient()
        zlib_ng = textwrap.dedent("""
            from conan import ConanFile
            class ZlibNG(ConanFile):
                name = "zlib-ng"
                version = "0.1"
                package_type = "static-library"
                def package_info(self):
                    self.cpp_info.libs = ["zlib"]
                    self.cpp_info.type = "static-library"
                    self.cpp_info.location = "lib/zlib.lib"
                    self.cpp_info.set_property("cmake_file_name", "ZLIB")
                    self.cpp_info.set_property("cmake_target_name", "ZLIB::ZLIB")
                    self.cpp_info.set_property("pkg_config_name", "ZLIB")
            """)
        openssl = textwrap.dedent("""
            from conan import ConanFile
            class openssl(ConanFile):
                name = "openssl"
                version = "0.1"
                package_type = "static-library"
                requires = "zlib/0.1"
                def package_info(self):
                    self.cpp_info.libs = ["crypto"]
                    self.cpp_info.type = "static-library"
                    self.cpp_info.location = "lib/crypto.lib"
                    self.cpp_info.requires = ["zlib::zlib"]
            """)
        zlib = '"zlib/0.1"' if diamond else ""
        conanfile = textwrap.dedent(f"""
            from conan import ConanFile
            class App(ConanFile):
                name = "app"
                version = "0.1"
                settings = "build_type", "arch"
                requires = "openssl/0.1", {zlib}
                package_type = "application"
                generators = "CMakeDeps", "PkgConfigDeps", "MSBuildDeps"
            """)
        profile = textwrap.dedent("""
            [settings]
            build_type = Release
            arch=x86_64

            [replace_requires]
            zlib/0.1: zlib-ng/0.1
            """)
        c.save({"zlibng/conanfile.py": zlib_ng,
                "openssl/conanfile.py": openssl,
                "app/conanfile.py": conanfile,
                "profile": profile})

        c.run("create zlibng")
        c.run("create openssl -pr=profile")
        c.run("install app -pr=profile -c tools.cmake.cmakedeps:new=will_break_next")
        assert "zlib/0.1: zlib-ng/0.1" in c.out

        pc_content = c.load("app/ZLIB.pc")
        assert 'Libs: -L"${libdir}" -lzlib' in pc_content
        pc_content = c.load("app/openssl.pc")
        assert 'Requires: ZLIB' in pc_content

        cmake = c.load("app/ZLIB-Targets-release.cmake")
        assert "add_library(ZLIB::ZLIB STATIC IMPORTED)" in cmake

        cmake = c.load("app/openssl-Targets-release.cmake")
        assert "find_dependency(ZLIB REQUIRED CONFIG)" in cmake
        assert "add_library(openssl::openssl STATIC IMPORTED)" in cmake
        assert "set_property(TARGET openssl::openssl APPEND PROPERTY INTERFACE_LINK_LIBRARIES\n" \
               '             "$<$<CONFIG:RELEASE>:ZLIB::ZLIB>")' in cmake

        # checking MSBuildDeps
        zlib_ng_props = c.load("app/conan_zlib-ng.props")
        assert 'Project="conan_zlib-ng_release_x64.props"' in zlib_ng_props
        props = c.load("app/conan_openssl_release_x64.props")
        assert "<Import Condition=\"'$(conan_zlib-ng_props_imported)' != 'True'\"" \
               " Project=\"conan_zlib-ng.props\"/>" in props

    @pytest.mark.parametrize("diamond", [True, False])
    def test_openssl_components(self, diamond):
        c = TestClient()
        zlib_ng = textwrap.dedent("""
            from conan import ConanFile
            class ZlibNG(ConanFile):
                name = "zlib-ng"
                version = "0.1"
                package_type = "static-library"
                def package_info(self):
                    self.cpp_info.libs = ["zlib"]
                    self.cpp_info.type = "static-library"
                    self.cpp_info.location = "lib/zlib.lib"
                    self.cpp_info.set_property("cmake_file_name", "ZLIB")
                    self.cpp_info.set_property("cmake_target_name", "ZLIB::ZLIB")
                    self.cpp_info.set_property("pkg_config_name", "ZLIB")
            """)
        openssl = textwrap.dedent("""
            from conan import ConanFile
            class openssl(ConanFile):
                name = "openssl"
                version = "0.1"
                package_type = "static-library"
                requires = "zlib/0.1"
                def package_info(self):
                    self.cpp_info.components["crypto"].libs = ["crypto"]
                    self.cpp_info.components["crypto"].type = "static-library"
                    self.cpp_info.components["crypto"].location = "lib/crypto.lib"
                    self.cpp_info.components["crypto"].requires = ["zlib::zlib"]
            """)
        zlib = '"zlib/0.1"' if diamond else ""
        conanfile = textwrap.dedent(f"""
            from conan import ConanFile
            class App(ConanFile):
                name = "app"
                version = "0.1"
                settings = "build_type", "arch"
                requires = "openssl/0.1", {zlib}
                package_type = "application"
                generators = "CMakeDeps", "PkgConfigDeps", "MSBuildDeps"
            """)
        profile = textwrap.dedent("""
            [settings]
            build_type = Release
            arch=x86_64

            [replace_requires]
            zlib/0.1: zlib-ng/0.1
            """)
        c.save({"zlibng/conanfile.py": zlib_ng,
                "openssl/conanfile.py": openssl,
                "app/conanfile.py": conanfile,
                "profile": profile})

        c.run("create zlibng")
        c.run("create openssl -pr=profile")
        c.run("install app -pr=profile -c tools.cmake.cmakedeps:new=will_break_next")
        assert "zlib/0.1: zlib-ng/0.1" in c.out

        pc_content = c.load("app/ZLIB.pc")
        assert 'Libs: -L"${libdir}" -lzlib' in pc_content
        pc_content = c.load("app/openssl-crypto.pc")
        assert 'Requires: ZLIB' in pc_content

        cmake = c.load("app/ZLIB-Targets-release.cmake")
        assert "add_library(ZLIB::ZLIB STATIC IMPORTED)" in cmake

        cmake = c.load("app/openssl-Targets-release.cmake")
        assert "find_dependency(ZLIB REQUIRED CONFIG)" in cmake
        assert "add_library(openssl::crypto STATIC IMPORTED)" in cmake
        assert "set_property(TARGET openssl::crypto APPEND PROPERTY INTERFACE_LINK_LIBRARIES\n" \
               '             "$<$<CONFIG:RELEASE>:ZLIB::ZLIB>")' in cmake

        # checking MSBuildDeps
        zlib_ng_props = c.load("app/conan_zlib-ng.props")
        assert 'Project="conan_zlib-ng_release_x64.props"' in zlib_ng_props

        props = c.load("app/conan_openssl_crypto_release_x64.props")
        assert "<Import Condition=\"'$(conan_zlib-ng_props_imported)' != 'True'\"" \
               " Project=\"conan_zlib-ng.props\"/>" in props

    @pytest.mark.parametrize("diamond", [True, False])
    @pytest.mark.parametrize("explicit_requires", [True, False])
    def test_zlib_components(self, diamond, explicit_requires):
        c = TestClient()
        zlib_ng = textwrap.dedent("""
            from conan import ConanFile
            class ZlibNG(ConanFile):
                name = "zlib-ng"
                version = "0.1"
                package_type = "static-library"
                def package_info(self):
                    self.cpp_info.components["myzlib"].libs = ["zlib"]
                    self.cpp_info.components["myzlib"].type = "static-library"
                    self.cpp_info.components["myzlib"].location = "lib/zlib.lib"
                    self.cpp_info.set_property("cmake_file_name", "ZLIB")
                    self.cpp_info.components["myzlib"].set_property("pkg_config_name", "ZLIB")
                    self.cpp_info.components["myzlib"].set_property("cmake_target_name",
                                                                    "ZLIB::ZLIB")
            """)
        openssl = textwrap.dedent(f"""
            from conan import ConanFile
            class openssl(ConanFile):
                name = "openssl"
                version = "0.1"
                package_type = "static-library"
                requires = "zlib/0.1"
                def package_info(self):
                    self.cpp_info.libs = ["crypto"]
                    self.cpp_info.type = "static-library"
                    self.cpp_info.location = "lib/crypto.lib"
                    if {explicit_requires}:
                        self.cpp_info.requires = ["zlib::zlib"]
            """)
        zlib = '"zlib/0.1"' if diamond else ""
        conanfile = textwrap.dedent(f"""
            from conan import ConanFile
            class App(ConanFile):
                name = "app"
                version = "0.1"
                settings = "build_type", "arch"
                requires = "openssl/0.1", {zlib}
                package_type = "application"
                generators = "CMakeDeps", "PkgConfigDeps", "MSBuildDeps"
            """)
        profile = textwrap.dedent("""
            [settings]
            build_type = Release
            arch = x86_64

            [replace_requires]
            zlib/0.1: zlib-ng/0.1
            """)
        c.save({"zlibng/conanfile.py": zlib_ng,
                "openssl/conanfile.py": openssl,
                "app/conanfile.py": conanfile,
                "profile": profile})

        c.run("create zlibng")
        c.run("create openssl -pr=profile")
        c.run("install app -pr=profile -c tools.cmake.cmakedeps:new=will_break_next")
        assert "zlib/0.1: zlib-ng/0.1" in c.out

        pc_content = c.load("app/zlib-ng.pc")
        assert 'Requires: ZLIB' in pc_content
        pc_content = c.load("app/ZLIB.pc")
        assert 'Libs: -L"${libdir}" -lzlib' in pc_content
        pc_content = c.load("app/openssl.pc")
        assert 'Requires: zlib-ng' in pc_content

        cmake = c.load("app/ZLIB-Targets-release.cmake")
        assert "add_library(ZLIB::ZLIB STATIC IMPORTED)" in cmake

        cmake = c.load("app/openssl-Targets-release.cmake")
        assert "find_dependency(ZLIB REQUIRED CONFIG)" in cmake
        assert "add_library(openssl::openssl STATIC IMPORTED)" in cmake
        # It should access the generic zlib-ng target
        assert "set_property(TARGET openssl::openssl APPEND PROPERTY INTERFACE_LINK_LIBRARIES\n" \
               '             "$<$<CONFIG:RELEASE>:zlib-ng::zlib-ng>")' in cmake

        # checking MSBuildDeps
        zlib_ng_props = c.load("app/conan_zlib-ng.props")
        assert "<Import Condition=\"'$(conan_zlib-ng_myzlib_props_imported)' != 'True'\" " \
               "Project=\"conan_zlib-ng_myzlib.props\"/" in zlib_ng_props

        props = c.load("app/conan_openssl_release_x64.props")
        assert "<Import Condition=\"'$(conan_zlib-ng_props_imported)' != 'True'\"" \
               " Project=\"conan_zlib-ng.props\"/>" in props

    @pytest.mark.parametrize("diamond", [True, False])
    @pytest.mark.parametrize("package_requires", [False, True])
    def test_both_components(self, diamond, package_requires):
        c = TestClient()
        zlib_ng = textwrap.dedent("""
            from conan import ConanFile
            class ZlibNG(ConanFile):
                name = "zlib-ng"
                version = "0.1"
                package_type = "static-library"
                def package_info(self):
                    self.cpp_info.components["myzlib"].libs = ["zlib"]
                    self.cpp_info.components["myzlib"].type = "static-library"
                    self.cpp_info.components["myzlib"].location = "lib/zlib.lib"
                    self.cpp_info.set_property("cmake_file_name", "ZLIB")
                    self.cpp_info.components["myzlib"].set_property("pkg_config_name", "ZLIB")
                    self.cpp_info.components["myzlib"].set_property("cmake_target_name",
                                                                    "ZLIB::ZLIB")
            """)
        openssl = textwrap.dedent(f"""
            from conan import ConanFile
            class openssl(ConanFile):
                name = "openssl"
                version = "0.1"
                package_type = "static-library"
                requires = "zlib/0.1"
                def package_info(self):
                    self.cpp_info.components["crypto"].libs = ["crypto"]
                    self.cpp_info.components["crypto"].type = "static-library"
                    self.cpp_info.components["crypto"].location = "lib/crypto.lib"
                    if {package_requires}:
                        self.cpp_info.components["crypto"].requires = ["zlib::zlib"]
                    else:
                        self.cpp_info.components["crypto"].requires = ["zlib::myzlib"]
            """)
        zlib = '"zlib/0.1"' if diamond else ""
        conanfile = textwrap.dedent(f"""
            from conan import ConanFile
            class App(ConanFile):
                name = "app"
                version = "0.1"
                settings = "build_type", "arch"
                requires = "openssl/0.1", {zlib}
                package_type = "application"
                generators = "CMakeDeps", "PkgConfigDeps", "MSBuildDeps"
            """)
        profile = textwrap.dedent("""
            [settings]
            build_type = Release
            arch = x86_64

            [replace_requires]
            zlib/0.1: zlib-ng/0.1
            """)
        c.save({"zlibng/conanfile.py": zlib_ng,
                "openssl/conanfile.py": openssl,
                "app/conanfile.py": conanfile,
                "profile": profile})

        c.run("create zlibng")
        c.run("create openssl -pr=profile")
        c.run("install app -pr=profile -c tools.cmake.cmakedeps:new=will_break_next")
        assert "zlib/0.1: zlib-ng/0.1" in c.out

        pc_content = c.load("app/zlib-ng.pc")
        assert 'Requires: ZLIB' in pc_content
        pc_content = c.load("app/ZLIB.pc")
        assert 'Libs: -L"${libdir}" -lzlib' in pc_content
        pc_content = c.load("app/openssl-crypto.pc")
        assert f'Requires: {"zlib-ng" if package_requires else "ZLIB"}' in pc_content

        cmake = c.load("app/ZLIB-Targets-release.cmake")
        assert "add_library(ZLIB::ZLIB STATIC IMPORTED)" in cmake

        cmake = c.load("app/openssl-Targets-release.cmake")
        assert "find_dependency(ZLIB REQUIRED CONFIG)" in cmake
        assert "add_library(openssl::crypto STATIC IMPORTED)" in cmake
        if package_requires:
            # The generic package requirement uses the package name zlib-ng
            assert "set_property(TARGET openssl::crypto APPEND PROPERTY INTERFACE_LINK_LIBRARIES\n" \
                   '             "$<$<CONFIG:RELEASE>:zlib-ng::zlib-ng>")' in cmake
        else:
            assert "set_property(TARGET openssl::crypto APPEND PROPERTY INTERFACE_LINK_LIBRARIES\n" \
                   '             "$<$<CONFIG:RELEASE>:ZLIB::ZLIB>")' in cmake

        # checking MSBuildDeps
        zlib_ng_props = c.load("app/conan_zlib-ng.props")
        assert "<Import Condition=\"'$(conan_zlib-ng_myzlib_props_imported)' != 'True'\" " \
               "Project=\"conan_zlib-ng_myzlib.props\"/" in zlib_ng_props

        props = c.load("app/conan_openssl_crypto_release_x64.props")
        if package_requires:
            assert "<Import Condition=\"'$(conan_zlib-ng_props_imported)' != 'True'\"" \
                   " Project=\"conan_zlib-ng.props\"/>" in props
        else:
            assert "<Import Condition=\"'$(conan_zlib-ng_myzlib_props_imported)' != 'True'\"" \
                   " Project=\"conan_zlib-ng_myzlib.props\"/>" in props


@pytest.mark.parametrize("require, pattern, alternative, expected", [
     # Version range as pattern
     # PINNED REQUIRE VERSION
     ("dep/1.0", "dep/[>=1.0 <2]", "dep/1.3", "dep/1.3"),
     ("dep/1.0", "dep/[>=1.5 <2]", "dep/1.3", False),
     # RANGE REQUIRE VERSION
     ("dep/[>=1.2 <2]", "dep/[>=1.0 <2]", "dep/1.3", "dep/1.3"),
     ("dep/[>=1.0 <1.5]", "dep/[>=1.2 <2]", "dep/1.3", "dep/1.3"),
     ("dep/[>=1.0 <1.5]", "dep/[>=1.5 <2]", "dep/1.3", False)
    ]
 )
def test_replace_requires_ranges(require, pattern, alternative, expected):
    c = TestClient(light=True)
    c.save({"dep/conanfile.py": GenConanfile("dep"),
            "app/conanfile.py": GenConanfile().with_requires(require),
            "profile": f"[replace_requires]\n{pattern}: {alternative}"})
    c.run("create dep --version=1.0")
    c.run("create dep --version=1.3")
    c.run("graph info app -pr=profile")
    if expected:
        assert "Replaced requires" in c.out
        assert f"{require}: {expected}" in c.out
    else:
        assert "Replaced requires" not in c.out
